# SEETM
PACKAGE_NAME = "seetm"
PACKAGE_NAME_PYPI = "seetm"
PACKAGE_VERSION = "0.0.1a2"
PACKAGE_VERSION_LONG = f'SEETM Version: {PACKAGE_VERSION}'
LANGUAGES_SUPPORTED = ['en', 'si']


class InterfaceType:
    CLI = "cli"
    INIT = "init"
    SERVER = "server"
    MAPPER = "map"
    EVALUATOR = "evaluate"
    EXTRACTOR = "extract"
    NONE = ""


class LoggingLevel:
    NOTSET = 0
    DEBUG = 10
    INFO = 20
    WARNING = 30
    ERROR = 40
    CRITICAL = 50
    QUIET = 60


class Validity:
    VALID = "valid"
    INVALID = "invalid"
    NOTSET = "n/a"


# SERVER
DEFAULT_PORT = 6067
DEFAULT_HOST_DEC = "0.0.0.0"
DEFAULT_HOST_LOCAL = "localhost"
SEETM_ASCII_LOGO = """
█▀ █▀▀ █▀▀ ▀█▀ █▀▄▀█
▄█ ██▄ ██▄ ░█░ █░▀░█
"""
SEETM_ASCII_LOGO_V2 = """
丂㠪㠪ㄒ爪
"""


class ServerEnv:
    STRICT_LOCAL = "strict_local"  # triggers server explanations without subprocess
    DEV = "dev"  # enables server debugging
    PROD = "prod"  # disables server debugging


class ServerConfigType:
    JSON = "json"
    NONE = "none"


# MAPPER
class MappingMethod:
    IPA = "ipa"
    RULE_BASED = "rule-based"
    TRANSLATE = "translate"


class MapperType:
    JSON = "json"
    DICT = "dict"
    QUIET = "quiet"


ENGLISH_TOKEN_PATTERN = r"^[a-zA-Z]+$"
EN_TO_PHONETICS_MAP_PATH = "./seetm_maps/en_to_phonetics_map.json"
TOKEN_TO_TOKEN_MAP_PATH = "./seetm_maps/token_to_token_map.json"

# SCAFFOLD
DEFAULT_INIT_SRC_DIR_NAME = "init_dir"
DEFAULT_INIT_CACHE_DIR_NAME = ".seetm_init_"
DEFAULT_INIT_FILES_TO_EXCLUDE = ['__pycache__', '__main__.py']
DEFAULT_INIT_DEST_DIR_NAME = "./"
RASA_DIRS_IN_SEETM_INIT = ["data"]
INVALID_DIR_NAME_CHARS = ['\\', '/', '<', '>', ':', '*', '?', '|']
ALLOWED_INIT_DIR_NAMES = [".", "./", "None", "none"]


class DestinationDirType:
    RASA = "rasa"
    SEETM = "seetm"
    EMPTY = "empty"
    VALID = "valid"
    INVALID = "invalid"


# RASA DATA FILES
DEFAULT_DATA_PATH = "./data"
YAML_EXTENSIONS = [".yaml", ".yml"]
DEFAULT_YAML_VERSION = "2.0"
DEFAULT_NLU_YAML_TAG = "nlu"
DEFAULT_VERSION_YAML_TAG = "version"
DEFAULT_NLU_INTENT_TAG = "intent"
DEFAULT_NLU_EXAMPLES_TAG = "examples"
DEFAULT_NLU_LOCATION_TAG = "location"


# EVAL DATA FILES
DEFAULT_EVALUATION_DATA_PATH = "./seetm_eval/"
DEFAULT_EVALUATION_RESULTS_PATH = "./seetm_eval/results"
DEFAULT_EMBEDDING_MODELS_PATH = "./seetm_eval/word2vec_models"

DEFAULT_EVAL_YAML_TAG = "eval"
DEFAULT_EVAL_TOKEN_TAG = "token"
DEFAULT_EVAL_TEST_TOKEN_TAG = "test_token"
DEFAULT_EVAL_MAPPED_TOKEN_TAG = "mapped_token"


class ExportableType:
    NLU = "nlu"
    EVAL = "eval"


class EvaluationLevel:
    TOKEN = "token"
    VECTOR = "vector"


# FILE IO
class FilePermission:
    READ = "r"
    WRITE = "w"


class Encoding:
    UTF8 = "utf8"


# TERMINAL
class TermColor:
    # source:
    # https://pkg.go.dev/github.com/whitedevops/colors

    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"
    END_C = "\033[0m"
    NONE_C = ""

    DEFAULT = "\033[39m"
    BLACK = "\033[30m"
    RED = "\033[31m"
    GREEN = "\033[32m"
    YELLOW = "\033[33m"
    BLUE = "\033[34m"
    MAGENTA = "\033[35m"
    CYAN = "\033[36m"
    LIGHTGRAY = "\033[37m"
    DARKGRAY = "\033[90m"
    LIGHTRED = "\033[91m"
    LIGHTGREEN = "\033[92m"
    LIGHTYELLOW = "\033[93m"
    LIGHTBLUE = "\033[94m"
    LIGHTMAGENTA = "\033[95m"
    LIGHTCYAN = "\033[96m"
    WHITE = "\033[97m"

    BDEFAULT = "\033[49m"
    BBLACK = "\033[40m"
    BRED = "\033[41m"
    BGREEN = "\033[42m"
    BYELLOW = "\033[43m"
    BBLUE = "\033[44m"
    BMAGENTA = "\033[45m"
    BCYAN = "\033[46m"
    BGRAY = "\033[47m"
    BDARKGRAY = "\033[100m"
    BLIGHTRED = "\033[101m"
    BLIGHTGREEN = "\033[102m"
    BLIGHTYELLOW = "\033[103m"
    BLIGHTBLUE = "\033[104m"
    BLIGHTMAGENTA = "\033[105m"
    BLIGHTCYAN = "\033[106m"
    BWHITE = "\033[107m"


# EVALUATOR
class Metrics:
    SEQUENCE_MATCHING = "sequence-matching"
    EDIT_DISTANCE = "edit-distance"


# CONFIGS
class Config:
    DEFAULT_CONFIG_PATH = "./seetm_config.yml"

    BASE_CONFIGS_KEY = "seetm_base_configs"
    SERVER_CONFIGS_KEY = "seetm_server_configs"

    DATA_PATH_KEY = "data_path"
    METHOD_KEY = "method"
    METRIC_KEY = "metric"
    CASE_SENSITIVE_KEY = "case_sensitive"

    HOST_KEY = "host"
    PORT_KEY = "port"

    VALID_MAIN_KEYS = ["seetm_base_configs", "seetm_server_configs"]
    VALID_BASE_KEYS = ["data_path", "method", "metric", "case_sensitive"]
    VALID_SERVER_KEYS = ["host", "port"]


class ConfigType:
    ALL = "all"
    BASE = "base"
    SERVER = "server"


# EXPORTABLE
DEFAULT_EXPORTABLE_PATH = "./seetm_exports/"
DEFAULT_TOKENIZER_PERSIST_PATH = "./seetm_exports/tokenizer_mappings.json"



import logging
import os
import re
from typing import Optional, List, Dict

from dime_xai.shared.constants import (
    RASA_288_MODEL_REGEX,
    DEFAULT_PERSIST_PATH,
    EXPLANATION_FILE_REGEX
)
from dime_xai.utils.io import (
    get_existing_toplevel_file_list,
    file_size
)

logger = logging.getLogger(__name__)


def model_statistics(request_data: Dict, reverse: bool) -> Optional[List]:
    models_path = request_data['models_path']
    models_list = get_existing_toplevel_file_list(
        dir_path=models_path,
    )
    model_sizes = file_size(
        file_path=[
            os.path.join(models_path, model_)
            for model_
            in models_list
            if re.match(RASA_288_MODEL_REGEX, model_)
        ],
        reverese=True if reverse else False,
    )
    model_sizes_list = [
        {
            "name": os.path.basename(model_),
            "size": str(round(size[0], 2)) + " " + str(size[1]),
        }
        for model_, size in model_sizes.items()]
    return model_sizes_list


def explanation_statistics(reverse: bool) -> Optional[List]:
    explanations_path = DEFAULT_PERSIST_PATH
    explanations_list = get_existing_toplevel_file_list(
        dir_path=explanations_path
    )
    explanations_list = [
        file_
        for file_
        in explanations_list
        if re.match(EXPLANATION_FILE_REGEX, file_)
    ]
    if reverse:
        explanations_list.reverse()
    return explanations_list



from typing import (
    Text,
    Optional,
    List,
    Union,
    Dict,
)
from copy import deepcopy
import regex

from seetm.utils.io import get_unique_list


def tokenize(instance: Text) -> Optional[List]:
    """
    Tokenizes a whitespace tokenize-able language.
    Tokenizing steps are the same as in Rasa
    WhitespaceTokenizer component

    Args:
        instance: whitespace tokenize-able string

    Returns:
        list of tokens or None
    """
    if not instance:
        return None

    # same logic from rasa tokenizer
    # to inherit the same tokenizing logic
    words = regex.sub(
        # there is a space or an end of a string after it
        r"[^\w#@&]+(?=\s|$)|"
        # there is a space or beginning of a string before it
        # not followed by a number
        r"(\s|^)[^\w#@&]+(?=[^0-9\s])|"
        # not in between numbers and not . or @ or & or - or #
        # e.g. 10'000.00 or blabla@gmail.com
        # and not url characters
        r"(?<=[^0-9\s])[^\w._~:/?#\[\]()@!$&*+,;=-]+(?=[^0-9\s])",
        " ",
        instance,
    ).split()

    tokens = [w for w in words if w]
    return tokens


def bag_of_words(
        instances: Union[Text, List],
        merge: bool = False
) -> Optional[Union[List, Dict]]:
    """
    Returns the bag of words of a string or
    a list of strings. Bag of words representation
    contains the unique list of words in a string or
    a list of strings

    Args:
        instances: a string or a list of strings
        merge: merges a list of strings as a single string is True.
        returns bag of words for each string in a list as a dictionary
        if set to False

    Returns:
        bag of words list for a string or a merged list of strings [List],
            bag of words list per each string in a list of strings [Dict], or None
    """
    instances_copy = deepcopy(instances)
    if not instances_copy:
        return None

    if merge and isinstance(instances_copy, List):
        instances_copy = ' '.join(instances_copy)

    if isinstance(instances_copy, List):
        instance_vocabulary = dict()
        for instance in instances_copy:
            bow = list()
            bow += tokenize(instance=instance)
            bow = sorted(get_unique_list(bow)) if bow else []
            instance_vocabulary[instance] = bow
        return instance_vocabulary

    elif isinstance(instances_copy, Text):
        vocabulary = list()
        vocabulary += tokenize(instance=instances_copy)
        vocabulary = sorted(get_unique_list(vocabulary)) if vocabulary else []
        return vocabulary


def get_token_count(token_list: List, token: Text) -> int:
    """
    Counts the number of instances of a token
    within a list of tokens passed

    Args:
        token_list: all tokens as a list
        token: token to get the number of instances

    Returns:
        number of instances as an int
    """
    return token_list.count(token)


def get_all_tokens(
        instances: Union[Text, List],
        merge: bool = False
) -> Optional[Union[List, Dict]]:
    """
    Returns all tokens present in a single string instance
    or a list of string instances

    Args:
        instances: single string instance or a list of strings
        merge: if True, mergers a list of strings passed as a single string

    Returns:
        for a single string, returns the list of tokens.
            for a list of unmerged strings, returns the list
            of tokens per each string instance as a dictionary
    """
    instances_copy = deepcopy(instances)
    if not instances_copy:
        return None

    if merge and isinstance(instances_copy, List):
        instances_copy = ' '.join(instances_copy)

    if isinstance(instances_copy, List):
        instance_vocabulary = dict()
        for instance in instances_copy:
            bow = list()
            bow += tokenize(instance=instance)
            instance_vocabulary[instance] = bow
        return instance_vocabulary

    elif isinstance(instances_copy, Text):
        vocabulary = list()
        vocabulary += tokenize(instance=instances_copy)
        return vocabulary


def remove_token(instance: Text, token: Text) -> Text:
    """
    Removes instances of a specified token
    from a single string instance

    Args:
        instance: string instance where the token should be removed from
        token: token to be removed

    Returns:
        token removed single string instance
    """
    instance = regex.sub(token + " ", "", instance)
    instance = regex.sub(" " + token, "", instance)
    return regex.sub(token, "", instance)


def remove_token_from_dataset(
        testing_data: Union[List, Dict],
        token: Text
) -> Union[List, Dict]:
    """
    Removes all instances of a specified
    token from the given dataset

    Args:
        testing_data: data instances as a list, or dictionary.
            If passed as dictionary, instances should be mentioned
            as a list under 'example' key under each intent/class

        token: token to be removed from the dataset

    Returns:
        token removed list or dictionary
    """
    testing_data_copy = deepcopy(testing_data)

    if isinstance(testing_data_copy, Dict):
        for intent, examples in testing_data_copy.items():
            testing_data_copy[intent] = [remove_token(example, token) for example in examples]
        return testing_data_copy
    else:
        for instance in testing_data_copy:
            instance['example'] = remove_token(instance=instance['example'], token=token)
        return testing_data_copy


def lowercase_list(instances_list: List) -> List:
    """
    lower-cases a list of strings

    Args:
        instances_list: list of strings to be lower-cased

    Returns:
        lower-cased list of strings
    """
    return [str.lower(instance) for instance in instances_list]


def order_dict(
        dict_to_order: Dict,
        order_by_key: bool = False,
        reverse: bool = False,
) -> Optional[Dict]:
    """
    Recursively orders a dictionary either
    by keys or values if comparable

    Args:
        dict_to_order: dictionary to be ordered
        order_by_key: orders by key if True, else orders by value
        reverse: orders in the ascending order if False,
            else orders in the descending order

    Returns:
        ordered dictionary, or None
    """
    if not dict_to_order:
        return None

    if order_by_key:
        index = 0
    else:
        index = 1

    ordered_dict = {
        k: v for k, v in sorted(
            dict_to_order.items(),
            key=lambda x: x[index],
            reverse=reverse)
    }

    for k, v in ordered_dict.items():
        if isinstance(v, Dict):
            ordered_dict[k] = order_dict(
                v,
                order_by_key=order_by_key,
                reverse=reverse
            )

    return ordered_dict



import logging
from typing import List, Text

import numpy

from seetm.shared.exceptions.core import (
    InvalidNumberOfTokensException,
)

logger = logging.getLogger(__name__)


def distance(tokens: List[Text], inspect: bool = False) -> int:
    if len(tokens) != 2:
        raise InvalidNumberOfTokensException()

    token_1 = tokens[0]
    token_2 = tokens[1]
    side_1 = len(token_1) + 1
    side_2 = len(token_2) + 1

    # initializing the distance matrix
    distance_matrix = numpy.zeros((side_1, side_2), int)
    for y in range(side_1):
        distance_matrix[y][0] = y
    for x in range(side_2):
        distance_matrix[0][x] = x

    # filling the distance matrix row by row
    for y in range(1, side_1):
        for x in range(1, side_2):
            insert_op = distance_matrix[y-1][x]
            delete_op = distance_matrix[y][x-1]
            replace_op = distance_matrix[y-1][x-1]

            if token_1[y-1] == token_2[x-1]:
                distance_matrix[y][x] = replace_op
            else:
                distance_matrix[y][x] = \
                    min([insert_op, delete_op, replace_op]) + 1

    # displaying the distance matrix if required
    if inspect:
        print(distance_matrix)
    return int(distance_matrix[side_1-1][side_2-1])


if __name__ == "__main__":
    logger.debug("Testing the Levenshtein distance...")
    print(distance(["dog", "ok"], inspect=True))



from seetm.shared.exceptions.base import SEETMException


class SEETMConfigException(SEETMException):
    pass


class InvalidInterfaceException(SEETMConfigException):
    pass


class InvalidConfigKeyException(SEETMConfigException):
    pass



import json
import logging
from copy import deepcopy
from typing import Text

import regex

from seetm.shared.constants import (
    Encoding,
    FilePermission,
    TOKEN_TO_TOKEN_MAP_PATH,
)

logger = logging.getLogger(__name__)


class TokenToTokenMapper:
    def __init__(self, case_sensitive: bool = True):
        self.case_sensitive = case_sensitive
        try:
            with open(
                    file=TOKEN_TO_TOKEN_MAP_PATH,
                    mode=FilePermission.READ,
                    encoding=Encoding.UTF8
            ) as token_to_token_map_file:
                self.token_to_token_map = json.load(token_to_token_map_file)
        except Exception as e:
            logger.debug(e)
            logger.warning("Failed to load token_to_token_map. "
                           "Rule based mappings will be ignored")
            self.token_to_token_map = {}

    def map(self, text: Text) -> Text:
        text_copy = deepcopy(text)
        if not self.case_sensitive:
            text_copy = str.lower(text_copy)

        for token_key, token_map in self.token_to_token_map.items():
            if not self.case_sensitive:
                token_map = [str.lower(token) for token in token_map]
            token_map = list(set(token_map))

            for pattern in token_map:
                text_copy = regex.sub(pattern, token_key, text_copy)

        return text_copy



import logging
from typing import List, Optional, Text, NoReturn, Dict

from seetm.shared.exceptions.io import EmptyNLUDatasetException
from seetm.utils.io import (
    get_rasa_testing_data,
    get_unique_list,
)

logger = logging.getLogger(__name__)


class RASANLUData:
    def __init__(
            self,
            data_path: Text = None,
    ):
        self.data_path = data_path
        self._initialize_nlu_data()

    def _initialize_nlu_data(self) -> NoReturn:
        logger.debug(f"Loading data using SEETM Testing data loader...")

        self._testing_data = get_rasa_testing_data(
            testing_data_dir=self.data_path,
            case_sensitive=True,
            preserve_location=True,
        )

        if not self._testing_data:
            raise EmptyNLUDatasetException()

        logger.info(f'Total number of intents: {self.get_intent_size()}')
        logger.info(f'Total number of data instances: {self.get_instance_size()}')

    def get_testing_data(
            self,
    ) -> Dict:
        testing_data = self._testing_data
        return testing_data

    def get_instances(
            self,
            intent: Text = None,
    ) -> Optional[List]:
        all_instances = list()
        if intent:
            if intent in self._testing_data:
                all_instances = self._testing_data[intent]
            else:
                logger.error("Could not find the given intent name in testing "
                             "data. Please input a valid intent name.")
        else:
            for key, value in self._testing_data.items():
                all_instances += value
        return get_unique_list(all_instances)

    def get_instance_size(
            self,
            intent: Text = None,
    ) -> Optional[int]:
        filtered_instances = self.get_instances(intent=intent)
        return len(filtered_instances)

    def get_intents(self) -> Optional[List]:
        if not self._testing_data:
            return []
        intents = list(self._testing_data.keys())
        return intents

    def get_intent_size(self) -> Optional[int]:
        if not self._testing_data:
            return 0
        return len(self._testing_data.keys())



from seetm.shared.exceptions.base import SEETMException


class SEETMServerException(SEETMException):
    pass


class ServerNotFoundException(SEETMServerException):
    pass


class ProcessTerminationException(SEETMServerException):
    pass


class InvalidProcessIDException(SEETMServerException):
    pass


class ProcessQueueException(SEETMServerException):
    pass


class ProcessQueuePushException(SEETMServerException):
    pass


class ProcessQueueUpdateException(SEETMServerException):
    pass


class ProcessQueuePullException(SEETMServerException):
    pass


class ProcessAlreadyExistsException(SEETMServerException):
    pass


class ProcessNotExistsException(SEETMServerException):
    pass


class MetadataRetrievalException(SEETMServerException):
    pass


class InvalidRequestIDException(SEETMServerException):
    pass


class ServerCacheException(SEETMServerException):
    pass


class ServerCachePushException(SEETMServerException):
    pass


class ServerCachePullException(SEETMServerException):
    pass


class ModelNotFoundException(SEETMServerException):
    pass


class ExplanationNotFoundException(SEETMServerException):
    pass


class InvalidExplanationSpecifiedException(SEETMServerException):
    pass


class InvalidServerConfigsException(SEETMServerException):
    pass


class ServerConfigsPersistException(SEETMServerException):
    pass


class InvalidConfigurationTypeSpecifiedException(SEETMServerException):
    pass


class CustomConfigsNotFoundException(SEETMServerException):
    pass



import logging
from copy import deepcopy
from typing import Text

import regex

from seetm.core.maps.ipa_to_sinhala_maps import (
    IPA_SI_VOWEL_KEYS,
    IPA_SI_VOWEL_SUFFIX_MAP,
    IPA_SI_CONSONANT_KEYS,
    IPA_SI_CONSONANT_MAP,
    EXACT_MATCH_KEYS,
    EXACT_MATCH_MAP,
)

logger = logging.getLogger(__name__)


class IPAToSinhalaMapper:
    @staticmethod
    def replace(text: Text, ipa_text: Text, map_phrases: bool = True) -> Text:
        word = str.lower(deepcopy(text))
        chars = str.lower(deepcopy(ipa_text))
        # STAGE 0
        # maps exact matches if set to true
        if map_phrases:
            # exact matches
            if word in EXACT_MATCH_KEYS:
                return EXACT_MATCH_MAP[word]

            # prefixes
            chars = regex.sub(r"^əˈwɔrd", "අවෝර්ඩ්", chars)
            chars = regex.sub(r"^ˌjunəˈvər", "යුනිවර්", chars)
            chars = regex.sub(r"^ækt", "ඇක්ට්", chars)
            chars = regex.sub(r"^ˈkɑmən", "කොමන්", chars)
            chars = regex.sub(r"^əˌsoʊʃi", "ඇසෝසි", chars)
            chars = regex.sub(r"^əˈkrɛdɪt", "ඇක්‍රෙඩිට්", chars)
            chars = regex.sub(r"^prəˈfɛʃən", "ප්‍රොෆෙශන්", chars)
            # terms
            chars = regex.sub(r"ˈnæʃənɑl", "නැශනල්", chars)
            chars = regex.sub(r"ˈnæʃənəl", "නැශනල්", chars)
            chars = regex.sub(r"^ˈpərsɪnəl$", "පර්සනල්", chars)
            chars = regex.sub(r"ʤɑb", "ජොබ්", chars)
            # suffixes
            chars = regex.sub(r"tut$", "ටියුට්", chars)
            chars = regex.sub(r"ˈtuʃən$", "ටියුෂන්", chars)
            chars = regex.sub(r"tud$", "ටියුඩ්", chars)
            chars = regex.sub(r"ɑləʤi$", "ඔලොජි", chars)
            chars = regex.sub(r"səti$", "සිටි", chars)
            chars = regex.sub(r"ˈeɪʃən$", "යේශන්", chars)
            chars = regex.sub(r"sətiz$", "සිටීස්", chars)
            chars = regex.sub(r"com$", "කොම්", chars)

        # STAGE 1
        # diphthongs
        chars = regex.sub(r"eɪ", "ඒ", chars)  # එයි
        chars = regex.sub(r"aɪ", "අයි", chars)
        chars = regex.sub(r"aʊ", "අව්", chars)
        chars = regex.sub(r"ɔɪ", "ඕ", chars)
        chars = regex.sub(r"oʊ", "ඕ", chars)  # ඔයි
        # monophthongs + stress
        chars = regex.sub(r"ˈæ", "ඇ", chars)
        chars = regex.sub(r"ˌæ", "ඇ", chars)
        # monophthongs
        chars = regex.sub(r"i", "ඊ", chars)
        chars = regex.sub(r"u", "ඌ", chars)
        chars = regex.sub(r"ɪ", "ඉ", chars)
        chars = regex.sub(r"ʊ", "උ", chars)
        chars = regex.sub(r"ə", "අ", chars)
        chars = regex.sub(r"ɛ", "එ", chars)
        chars = regex.sub(r"ɝ", "අර්", chars)
        chars = regex.sub(r"æ", "ඈ", chars)
        chars = regex.sub(r"ɑ", "ආ", chars)
        # consonants
        chars = regex.sub(r"tʃ", "ච්", chars)
        chars = regex.sub(r"dʒ", "ජ්", chars)
        chars = regex.sub(r"p", "ප්", chars)
        chars = regex.sub(r"b", "බ්", chars)
        chars = regex.sub(r"t", "ට්", chars)
        chars = regex.sub(r"d", "ඩ්", chars)
        chars = regex.sub(r"k", "ක්", chars)
        chars = regex.sub(r"g", "ග්", chars)
        chars = regex.sub(r"θ", "ත්", chars)
        chars = regex.sub(r"ð", "ද්", chars)
        chars = regex.sub(r"f", "ෆ්", chars)
        chars = regex.sub(r"v", "ව්", chars)
        chars = regex.sub(r"s", "ස්", chars)
        chars = regex.sub(r"z", "ස්", chars)
        chars = regex.sub(r"ʃ", "ශ්", chars)
        chars = regex.sub(r"ʒ", "ෂ්", chars)
        chars = regex.sub(r"w", "ව්", chars)
        chars = regex.sub(r"ɫ", "ල්", chars)
        chars = regex.sub(r"m", "ම්", chars)
        chars = regex.sub(r"n", "න්", chars)
        chars = regex.sub(r"ŋ", "න්ග්", chars)
        chars = regex.sub(r"ɹ", "ර්", chars)
        chars = regex.sub(r"j", "ය්", chars)
        chars = regex.sub(r"h", "හ්", chars)
        # modern
        chars = regex.sub(r"l", "ල්", chars)
        chars = regex.sub(r"r", "ර්", chars)
        chars = regex.sub(r"ʤ", "ජ්", chars)
        chars = regex.sub(r"ɔ", "අව්", chars)

        # STAGE 2
        # stress (primary(') or secondary(,))
        # remove remaining stress marks
        chars = regex.sub(r"ˈ", "", chars)
        chars = regex.sub(r"ˌ", "", chars)

        # STAGE 3
        # combine consonants and vowels that
        # appear next to each other
        for consonant in IPA_SI_CONSONANT_KEYS:
            for vowel in IPA_SI_VOWEL_KEYS:
                # consonant + ර් + vowel
                pattern_before = consonant + 'ර්' + vowel
                pattern_after = IPA_SI_CONSONANT_MAP[consonant] + "්‍ර" + IPA_SI_VOWEL_SUFFIX_MAP[vowel]
                chars = regex.sub(pattern_before, pattern_after, chars)

                # consonant + vowel
                pattern_before = consonant + vowel
                pattern_after = IPA_SI_CONSONANT_MAP[consonant] + IPA_SI_VOWEL_SUFFIX_MAP[vowel]
                chars = regex.sub(pattern_before, pattern_after, chars)
        return chars



import json
import logging
import os
from typing import Union, Text, NoReturn, List, Dict

from seetm.shared.constants import (
    DEFAULT_EXPORTABLE_PATH,
    Encoding,
    FilePermission,
)
from seetm.shared.exceptions.core import (
    ExportableInitializationException, ExportablePersistException,
)
from seetm.utils.io import get_timestamp_str

logger = logging.getLogger(__name__)


class Exportable:
    def __init__(self, exportable: Union[List, Text]) -> None:
        if isinstance(exportable, str):
            self.name = exportable
            self.content = self._initialize(exportable=exportable)
        else:
            self.name = f"seetm_export_{get_timestamp_str()}.json"
            self.content = exportable

    @staticmethod
    def _initialize(exportable: Text) -> Dict:
        try:
            if len(os.path.split(exportable)) != 1:
                raise ExportableInitializationException()

            with open(
                    os.path.join(DEFAULT_EXPORTABLE_PATH, exportable),
                    encoding=Encoding.UTF8,
                    mode=FilePermission.READ
            ) as exportable_file:
                exportable_content = json.load(exportable_file)

            return exportable_content
        except Exception:
            logger.error(f"Failed to initialize the specified "
                         f"SEETM exportable {exportable}")
            raise ExportableInitializationException()

    def persist(self, indent: int = 4) -> NoReturn:
        try:
            with open(
                    os.path.join(DEFAULT_EXPORTABLE_PATH, self.name),
                    encoding=Encoding.UTF8,
                    mode=FilePermission.WRITE
            ) as exportable_file:
                json.dump(
                    self.content,
                    exportable_file,
                    ensure_ascii=False,
                    indent=indent
                )
                logger.debug(f"Exportable {self.name} "
                             f"was persisted")
        except Exception:
            logger.error(f"Failed to persist the specified "
                         f"SEETM exportable {self.name}")
            raise ExportablePersistException()

    def inspect(self):
        print(json.dumps(self.content, ensure_ascii=False, indent=4))



import logging
import os
from typing import Dict

from waitress import serve as waitress_serve

from seetm.server import create_app
from seetm.shared.constants import (
    DEFAULT_PORT,
    DEFAULT_HOST_LOCAL,
    SEETM_ASCII_LOGO,
    ServerEnv, Config,
)
from seetm.shared.exceptions.base import (
    SEETMException,
)
from seetm.shared.exceptions.core import (
    SEETMCoreException
)
from seetm.shared.exceptions.server import (
    SEETMServerException,
    ServerNotFoundException,
)

logger = logging.getLogger(__name__)


class SEETMServer:
    def __init__(
            self,
            configs: Dict,
            debug_mode: bool = False,
    ):
        self.configs = configs
        self.host = configs[Config.SERVER_CONFIGS_KEY][Config.HOST_KEY] or DEFAULT_HOST_LOCAL
        self.port = configs[Config.SERVER_CONFIGS_KEY][Config.PORT_KEY] or DEFAULT_PORT
        self.data_path = configs[Config.BASE_CONFIGS_KEY][Config.DATA_PATH_KEY]
        self.method = configs[Config.BASE_CONFIGS_KEY][Config.METHOD_KEY]
        self.debug_mode = debug_mode

    def run(self) -> None:
        logger.info(f"Starting SEETM server at http://{self.host}:{self.port}/")
        try:
            configs = self.configs
            app_config = {
                "SEETM": configs,
                "APP_THEME": os.environ.get("APP_THEME") or "dark",
                "APP_ENV": os.environ.get("APP_ENV") or "prod",
                "SINHALA_ENABLED": os.environ.get("SINHALA_ENABLED") or True,
            }

            if self.debug_mode:
                logger.warning("Deploying SEETM Server in development mode...")
                os.environ["APP_ENV"] = ServerEnv.DEV
                app = create_app(configs=app_config)
                app.run(
                    host=self.host,
                    port=self.port,
                    debug=self.debug_mode
                )
            else:
                logger.info("Deploying SEETM Server in production mode...")
                print(SEETM_ASCII_LOGO)
                waitress_serve(
                    create_app(configs=app_config),
                    host=self.host,
                    port=self.port
                )

                # # Run as a shell command if required
                # import subprocess
                # subprocess.run(["waitress-serve", f"--host={self.host}",
                #                 f"--port={self.port}", "seetm.server.seetm_server:run"])

        except ServerNotFoundException:
            logger.exception(f"An unknown exception occurred while invoking the SEETMServer")
        except SEETMCoreException:
            logger.exception(f"Core::SEETMServer")
        except SEETMServerException:
            logger.exception(f"Specific::SEETMServer")
        except SEETMException:
            logger.exception(f"Base::SEETMServer")
        except KeyboardInterrupt:
            logger.info(f"Gracefully terminating SEETM Server...")
            exit()
        except OSError:
            logger.exception(f"Possible permission exception while starting the SEETMServer")
        except Exception as e:
            logger.exception(f"Base::broad::SEETMServer. more info: {e}")
        return



import json
import logging
from copy import deepcopy
from typing import Text

from seetm.shared.constants import (
    Encoding,
    FilePermission,
    EN_TO_PHONETICS_MAP_PATH,
)

logger = logging.getLogger(__name__)


class EnglishToPhoneticsMapper:
    def __init__(self):
        try:
            with open(
                    file=EN_TO_PHONETICS_MAP_PATH,
                    mode=FilePermission.READ,
                    encoding=Encoding.UTF8
            ) as en_to_phonetics_map_file:
                self.en_to_phonetics_map = json.load(en_to_phonetics_map_file)
        except Exception as e:
            logger.debug(e)
            logger.warning("Failed to load en_to_phonetics_map.json. External mappings will be ignored")
            self.en_to_phonetics_map = {}

    def map(self, token: Text) -> Text:
        token_copy = str.lower(deepcopy(token))
        if token_copy in self.en_to_phonetics_map:
            mapped_token = self.en_to_phonetics_map[token_copy]
            return mapped_token if mapped_token else token
        else:
            return token




import logging
from typing import Text, Union, Tuple, List

import eng_to_ipa as ipa
import regex

from seetm.core.mappers.english_to_phonetics_mapper import (
    EnglishToPhoneticsMapper,
)
from seetm.core.mappers.ipa_to_sinhala_mapper import (
    IPAToSinhalaMapper,
)
from seetm.core.mappers.token_to_token_mapper import (
    TokenToTokenMapper,
)
from seetm.shared.constants import (
    MappingMethod,
    ENGLISH_TOKEN_PATTERN,
)
from seetm.shared.exceptions.core import (
    InvalidDataInstanceException,
    MapperException, InvalidMappingMethodException,
)
from seetm.utils.text_preprocessing import tokenize

logger = logging.getLogger(__name__)
EN_PATTERN = regex.compile(ENGLISH_TOKEN_PATTERN)


class TokenMapper:
    def __init__(
            self,
            method: Text = MappingMethod.RULE_BASED,
            case_sensitive: bool = True,
            export: bool = False,
            quiet_mode: bool = False,
    ) -> None:
        self.method = method
        self.case_sensitive = case_sensitive
        self.export = export
        self.quiet_mode = quiet_mode
        logger.debug('Initialized the SEETM mapper')

    def map_and_persist(self):
        raise NotImplementedError()

    def map(
            self,
            data_instance: Text = None,
            token_list: bool = False,
    ) -> Union[Text, Tuple[Text, List]]:
        try:
            if self.method not in [
                MappingMethod.IPA,
                MappingMethod.RULE_BASED
            ]:
                raise NotImplementedError()

            if not data_instance:
                raise InvalidDataInstanceException()

            if self.method == MappingMethod.IPA:
                en_to_phonetics_mapper = EnglishToPhoneticsMapper()

                # tokenize
                tokens = tokenize(data_instance)

                # replacing from user-provided pronunciations
                tokens = [
                    en_to_phonetics_mapper.map(token=token)
                    if len(EN_PATTERN.findall(token)) == 1 else token
                    for token in tokens
                ]

                # converting the rest of remaining english
                # words to IPA representation
                mapped_tokens = [
                    {
                        "token": t,
                        "status": ipa.isin_cmu(t),
                        "ipa": ipa.convert(t) if ipa.isin_cmu(t) else t
                    }
                    for t
                    in tokens
                    if t
                ]

                # converting the IPA representation to
                # the appropriate sinhala representation
                mapped_token_list = list()
                for token in mapped_tokens:
                    if token["status"]:
                        mapped_token = IPAToSinhalaMapper.replace(
                            text=token["token"],
                            ipa_text=token["ipa"],
                            map_phrases=True
                        )
                    else:
                        mapped_token = token["token"]
                    token["mapped_token"] = mapped_token

                    if token_list:
                        mapped_token_list.append({
                            "token": token["token"],
                            "mapped_token": mapped_token,
                        })

                # return the combined tokens
                mapped_text = " ".join(
                    [token["mapped_token"]
                     for token
                     in mapped_tokens]
                ) if mapped_tokens else ""

                return mapped_text, mapped_token_list

            elif self.method == MappingMethod.RULE_BASED:
                token_to_token_mapper = TokenToTokenMapper(case_sensitive = self.case_sensitive)
                mapped_text = token_to_token_mapper.map(text=data_instance)

                return mapped_text
            else:
                raise InvalidMappingMethodException()
        except Exception as e:
            raise MapperException(e)



import argparse
import logging
import os
import sys
from typing import NoReturn, Text

from dotenv import load_dotenv

from seetm.cli.seetm_cli import SEETMCLIMapper
from seetm.core.evaluator import SEETMEvaluator
from seetm.core.extractor import SEETMExtractor
from seetm.server.seetm_server import SEETMServer
from seetm.shared.constants import (
    InterfaceType,
    TermColor,
    MappingMethod,
    LoggingLevel,
    Metrics,
    Config,
    PACKAGE_VERSION_LONG,
    EvaluationLevel,
)
from seetm.shared.exceptions.config import (
    InvalidInterfaceException,
    InvalidConfigKeyException,
)
from seetm.shared.exceptions.core import (
    ZeroMappedTokensException,
    EvaluationDatasetExportException,
    ZeroValidTokensException,
    OutDatedEvaluationDatasetException
)
from seetm.utils.config import get_init_configs
from seetm.utils.io import (
    set_cli_color,
    dir_exists,
    update_sys_path,
)
from seetm.utils.scaffold import SEETMInit
from seetm.utils.seetm_logging_formatter import (
    SEETMLoggingFormatter,
    MaxLevelFilter,
)

load_dotenv()
logger = logging.getLogger()
update_sys_path(os.getcwd())

formatter = SEETMLoggingFormatter(format_str='%(asctime)s\t%(levelname)s\t%(name)s - %(message)s')
logging_out = logging.StreamHandler(sys.stdout)
logging_err = logging.StreamHandler(sys.stderr)
logging_out.setFormatter(formatter)
logging_err.setFormatter(formatter)
logging_out.addFilter(MaxLevelFilter(logging.WARNING))
logging_out.setLevel(logging.DEBUG)
logging_err.setLevel(logging.WARNING)
logger.addHandler(logging_out)
logger.addHandler(logging_err)
logger.setLevel(level=logging.INFO)


def create_argument_parser():
    parser = argparse.ArgumentParser(prog="seetm", description="starts SEETM CLI")
    subparsers = parser.add_subparsers(help='desired SEETM interface to run [cli/server]', dest="subparser_name")

    parser.add_argument(
        "-v",
        "--version",
        action='version',
        version=PACKAGE_VERSION_LONG,
        help="prints the SEETM version info.",
    )

    parser_init = subparsers.add_parser(
        name="init",
        help='create init dir structure for a new mapping project.'
    )
    parser_init.add_argument(
        "--debug",
        action="store_true",
        help="sets the SEETM logging level to debug mode from info.",
    )
    parser_init.add_argument(
        "--quiet",
        action="store_true",
        help="initializes a starter SEETM project without prompting the user for a project configs.",
    )

    parser_server = subparsers.add_parser(
        name="server",
        help='run SEETM server, a web-based mapping tool.'
    )
    parser_server.add_argument(
        "-p",
        "--port",
        type=int,
        help="the port to start the SEETM server at.",
    )
    parser_server.add_argument(
        "-c",
        "--config",
        type=str,
        default=Config.DEFAULT_CONFIG_PATH,
        help="custom SEETM config file.",
    )
    parser_server.add_argument(
        "-m",
        "--method",
        type=str,
        choices=[MappingMethod.IPA, MappingMethod.RULE_BASED, MappingMethod.TRANSLATE],
        help="method used to map the instance/token.",
    )
    parser_server.add_argument(
        "--debug",
        action="store_true",
        help="sets the logging level to debug mode from info and flask server debug mode to True.",
    )

    parser_mapper = subparsers.add_parser(
        name="map",
        help='run SEETM CLI mapper, a terminal-based mapper tool.'
    )
    parser_mapper.add_argument(
        "-i",
        "--instance",
        type=str,
        default=None,
        help="data instance to generate mapping.",
    )
    parser_mapper.add_argument(
        "-c",
        "--config",
        type=str,
        default=Config.DEFAULT_CONFIG_PATH,
        help="custom SEETM config file.",
    )
    parser_mapper.add_argument(
        "-m",
        "--method",
        type=str,
        choices=[MappingMethod.IPA, MappingMethod.RULE_BASED],
        help="method used to map the instance/token.",
    )
    parser_mapper.add_argument(
        "--debug",
        action="store_true",
        help="sets the SEETM logging level to debug mode from info.",
    )
    parser_mapper.add_argument(
        "--no-persist",
        action="store_false",
        dest="persist",
        help="states whether to store mapped instance or not.",
    )
    parser_mapper.add_argument(
        "--no-case",
        action="store_false",
        dest="case",
        help="states whether to store mapped instance or not.",
    )
    parser_mapper.add_argument(
        "--quiet",
        action="store_true",
        help="sets the SEETM logging level to off.",
    )

    parser_evaluator = subparsers.add_parser(
        name="evaluate",
        help='run SEETM CLI evaluator.'
    )
    parser_evaluator.add_argument(
        "-c",
        "--config",
        type=str,
        default=Config.DEFAULT_CONFIG_PATH,
        help="custom SEETM config file.",
    )
    parser_evaluator.add_argument(
        "-m",
        "--method",
        type=str,
        choices=[MappingMethod.IPA, MappingMethod.RULE_BASED],
        help="method used to map the instance/token.",
    )
    parser_evaluator.add_argument(
        "--metric",
        type=str,
        choices=[Metrics.EDIT_DISTANCE, Metrics.SEQUENCE_MATCHING],
        help="metric to evaluate mapped tokens or instances.",
    )
    parser_evaluator.add_argument(
        "-l",
        "--level",
        type=str,
        choices=[EvaluationLevel.TOKEN, EvaluationLevel.VECTOR],
        required=True,
        help="sets the evaluation level to token or instance level.",
    )
    parser_evaluator.add_argument(
        "-e",
        "--epochs",
        type=int,
        default=10,
        help="sets the number of epochs to train the word2vec models.",
    )
    parser_evaluator.add_argument(
        "--debug",
        action="store_true",
        help="sets the SEETM logging level to debug mode from info.",
    )
    parser_evaluator.add_argument(
        "--no-persist",
        action="store_false",
        dest="persist",
        help="states whether to store mapped instance or not.",
    )
    parser_evaluator.add_argument(
        "--quiet",
        action="store_true",
        help="sets the SEETM logging level to off.",
    )

    parser_extractor = subparsers.add_parser(
        name="extract",
        help='run SEETM CLI token extractor.'
    )
    parser_extractor.add_argument(
        "-d",
        "--data-path",
        type=str,
        default=None,
        help="data path to find NLU files at.",
    )
    parser_extractor.add_argument(
        "-c",
        "--config",
        type=str,
        default=Config.DEFAULT_CONFIG_PATH,
        help="custom SEETM config file.",
    )
    parser_extractor.add_argument(
        "--clean",
        action="store_true",
        help="informs the extractor to remove existing eval data and maps.",
    )
    parser_extractor.add_argument(
        "--debug",
        action="store_true",
        help="sets the SEETM logging level to debug mode from info.",
    )
    parser_extractor.add_argument(
        "--no-persist",
        action="store_false",
        dest="persist",
        help="states whether to store mapped instance or not.",
    )
    parser_extractor.add_argument(
        "--quiet",
        action="store_true",
        help="sets the SEETM logging level to off.",
    )
    return parser


def _set_logging_level(level: Text = LoggingLevel.INFO) -> NoReturn:
    if level == LoggingLevel.NOTSET:
        logger.setLevel(level=logging.NOTSET)
    elif level == LoggingLevel.DEBUG:
        logger.setLevel(level=logging.DEBUG)
    elif level == LoggingLevel.INFO:
        logger.setLevel(level=logging.INFO)
    elif level == LoggingLevel.WARNING:
        logger.setLevel(level=logging.WARNING)
    elif level == LoggingLevel.ERROR:
        logger.setLevel(level=logging.ERROR)
    elif level == LoggingLevel.CRITICAL:
        logger.setLevel(level=logging.CRITICAL)
    elif level == LoggingLevel.QUIET:
        logging.disable(level=logging.CRITICAL)
    else:
        logger.setLevel(level=logging.INFO)


def run_seetm_cli() -> NoReturn:
    try:
        logger.debug("Running main SEETM CLI.")
        arg_parser = create_argument_parser()
        cmdline_args = arg_parser.parse_args()
        interface = cmdline_args.subparser_name

        if not interface:
            arg_parser.print_help()
            logger.error("Please specify a valid positional arg out of \'map\',\'init\', \'server\', \'evaluate\', "
                         "and \'extract\' to use SEETM CLI.")
            return

        if str.lower(interface) == InterfaceType.INIT:
            quiet = cmdline_args.quiet
            debug_mode = cmdline_args.debug

            if debug_mode:
                _set_logging_level(level=LoggingLevel.DEBUG)
            else:
                _set_logging_level(level=LoggingLevel.INFO)

            try:
                if not quiet:
                    print(set_cli_color(
                        text_content="👋🏽 Hi there! Welcome to SEETM.",
                        color=TermColor.LIGHTGREEN)
                    )
                    dest_dir = input(set_cli_color(
                        text_content="In which directory do you want to "
                                     "initialize SEETM? [Default: Current "
                                     "Directory]: ",
                        color=TermColor.LIGHTGREEN
                    ))
                else:
                    dest_dir = "."

                if dest_dir and not dir_exists(dir_path=dest_dir):
                    logger.error("Directory name or path should be a "
                                 "valid existing directory")
                    return

                seetm_init = SEETMInit()
                seetm_init.build_scaffold(dest_path=dest_dir)
            except KeyboardInterrupt:
                logger.error("Gracefully terminating SEETM init...")

        elif str.lower(interface) == InterfaceType.SERVER:
            port = cmdline_args.port
            config = cmdline_args.config
            method = cmdline_args.method
            debug_mode = cmdline_args.debug

            if debug_mode:
                _set_logging_level(level=LoggingLevel.DEBUG)
            else:
                _set_logging_level(level=LoggingLevel.INFO)

            configs = get_init_configs(
                config_path=config,
                port=port,
                method=method,
                interface=InterfaceType.SERVER,
            )

            seetm_server = SEETMServer(
                configs=configs,
                debug_mode=debug_mode or False,
            )
            seetm_server.run()

        elif str.lower(interface) == InterfaceType.MAPPER:
            instance = cmdline_args.instance
            method = cmdline_args.method
            config = cmdline_args.config
            persist = cmdline_args.persist
            case = cmdline_args.case
            debug_mode = cmdline_args.debug
            quiet_mode = cmdline_args.quiet

            if debug_mode:
                _set_logging_level(level=LoggingLevel.DEBUG)
            elif quiet_mode:
                _set_logging_level(level=LoggingLevel.QUIET)
            else:
                _set_logging_level(level=LoggingLevel.INFO)

            configs = get_init_configs(
                config_path=config,
                instance=instance,
                method=method,
                case=case,
                interface=InterfaceType.MAPPER,
            )
            print(configs)
            exit()
            seetm_cli_mapper = SEETMCLIMapper(
                configs=configs,
                persist=persist,
                quiet_mode=True if quiet_mode else False,
            )
            seetm_cli_mapper.run()

        elif str.lower(interface) == InterfaceType.EVALUATOR:
            metric = cmdline_args.metric
            method = cmdline_args.method
            level = cmdline_args.level
            config = cmdline_args.config
            epochs = cmdline_args.epochs
            persist = cmdline_args.persist
            debug_mode = cmdline_args.debug
            quiet_mode = cmdline_args.quiet

            if debug_mode:
                _set_logging_level(level=LoggingLevel.DEBUG)
            elif quiet_mode:
                _set_logging_level(level=LoggingLevel.QUIET)
            else:
                _set_logging_level(level=LoggingLevel.INFO)

            if metric == Metrics.SEQUENCE_MATCHING:
                logger.warning("sequence-matching metric depends on the order of the "
                               "tokens provided (test_instance, mapped_instance). "
                               "Use edit-distance as it is more precise.")

            configs = get_init_configs(
                config_path=config,
                method=method,
                metric=metric,
                level=level,
                interface=InterfaceType.EVALUATOR,
            )
            seetm_evaluator = SEETMEvaluator(
                configs=configs,
                level=level,
                epochs=epochs,
                persist=persist,
                quiet_mode=True if quiet_mode else False,
            )
            seetm_evaluator.evaluate()

        elif str.lower(interface) == InterfaceType.EXTRACTOR:
            data_path = cmdline_args.data_path
            config = cmdline_args.config
            persist = cmdline_args.persist
            clean = cmdline_args.clean
            debug_mode = cmdline_args.debug
            quiet_mode = cmdline_args.quiet

            if debug_mode:
                _set_logging_level(level=LoggingLevel.DEBUG)
            elif quiet_mode:
                _set_logging_level(level=LoggingLevel.QUIET)
            else:
                _set_logging_level(level=LoggingLevel.INFO)

            configs = get_init_configs(
                config_path=config,
                data_path=data_path,
                interface=InterfaceType.EXTRACTOR,
            )
            seetm_extractor = SEETMExtractor(
                configs=configs,
                persist=persist,
                clean=clean,
                quiet_mode=True if quiet_mode else False,
            )
            seetm_extractor.extract()
        else:
            logger.error('One or more incorrect CLI arguments detected. '
                         'Refer "seetm -h" to view allowed arguments')
            return
    except InvalidInterfaceException as e:
        logger.error(f"Invalid interface specified.")
        logger.debug(e) if e else None
    except InvalidConfigKeyException as e:
        logger.error(f"Found invalid config keys.")
        logger.debug(e) if e else None
    except KeyboardInterrupt:
        logger.info(f"Gracefully terminating SEETM CLI...")

    # new
    except ZeroMappedTokensException as e:
        logger.error(f"Non of the tokens were mapped. Evaluation is terminated. {e}")
    except EvaluationDatasetExportException as e:
        logger.error(f"Exception occurred while exporting the evaluation dataset. {e}")

    except ZeroValidTokensException:
        logger.error(f"Could not find any valid evaluation token pairs. "
                     f"Use `seetm extract` to generate an up-to-date evaluation "
                     f"dataset or manually map them and retry")
    except OutDatedEvaluationDatasetException:
        logger.error(f"Found out-dated evaluation token pairs. "
                     f"Use `seetm extract` to generate an up-to-date evaluation "
                     f"dataset or manually remove them and retry")

    #     # Training data exceptions
    #     except EvalFileNotFoundException:
    #     logger.error("Could not find the eval YAML files")
    #     exit("data not found error") if self.quiet_mode else exit()
    # except YAMLFormatException:
    #     logger.error("Invalid YAML file format in NLU files")
    #     exit("invalid yaml format error") if self.quiet_mode else exit()
    # except InvalidNLUDatasetException:
    #     logger.error("Could not find the data dir specified")
    #     exit("invalid data path error") if self.quiet_mode else exit()
    # except EmptyNLUDatasetException as e:
    #     logger.error(e)
    #     exit("empty nlu data error") if self.quiet_mode else exit()
    # except InvalidFileExtensionException as e:
    #     logger.error(e)
    #     exit("file ext error") if self.quiet_mode else exit()
    # except InvalidDataInstanceException as e:
    #     logger.error(e)
    #     exit("data instance error") if self.quiet_mode else exit()
    # except ExportableInitializationException as e:
    #     logger.error(e)
    #     exit("data instance error") if self.quiet_mode else exit()
    # # SEETM Base exceptions
    # except MapperException as e:
    #     logger.error(e)
    #     exit("mapper error") if self.quiet_mode else exit()
    # except SEETMCoreException as e:
    #     logger.error(f"Unknown Core Exception {e}")
    #     exit("core error") if self.quiet_mode else exit()
    # except SEETMIOException as e:
    #     logger.error(f"Unknown IO Exception {e}")
    #     exit("io error") if self.quiet_mode else exit()
    # except SEETMException as e:
    #     logger.error(f"Unknown Base Exception {e}")
    #     exit("base error") if self.quiet_mode else exit()
    # # Default exceptions
    # except KeyboardInterrupt:
    #     logger.info(f"Gracefully terminating SEETM CLI Mapper...")
    #     exit("keyboard error") if self.quiet_mode else exit()
    # except MemoryError as e:
    #     logger.info(f"Out of memory exception occurred "
    #                 f"while loading the model. {e}")
    #     exit("memory error") if self.quiet_mode else exit()
    # except OSError as e:
    #     logger.info(f"OSError exception occurred "
    #                 f"while loading the model. {e}")
    #     exit("os error") if self.quiet_mode else exit()
    # # except Exception as e:
    # #     logger.error(f"Unknown CLI Exception. {e}")
    # #     exit("generic error") if self.quiet_mode else exit()
    # #
    #     # Training data exceptions
    #     except NLUFileNotFoundException:
    #     logger.error("Could not find the NLU yaml files")
    #     exit("data not found error") if self.quiet_mode else exit()
    # except YAMLFormatException:
    #     logger.error("Invalid YAML file format in NLU files")
    #     exit("invalid yaml format error") if self.quiet_mode else exit()
    # except InvalidNLUDatasetException:
    #     logger.error("Could not find the data dir specified")
    #     exit("invalid data path error") if self.quiet_mode else exit()
    # except EmptyNLUDatasetException as e:
    #     logger.error(e)
    #     exit("empty nlu data error") if self.quiet_mode else exit()
    # except InvalidFileExtensionException as e:
    #     logger.error(e)
    #     exit("file ext error") if self.quiet_mode else exit()
    # except InvalidDataInstanceException as e:
    #     logger.error(e)
    #     exit("data instance error") if self.quiet_mode else exit()
    # except ExportableInitializationException as e:
    #     logger.error(e)
    #     exit("data instance error") if self.quiet_mode else exit()
    # # SEETM Base exceptions
    # except MapperException as e:
    #     logger.error(e)
    #     exit("mapper error") if self.quiet_mode else exit()
    # except SEETMCoreException as e:
    #     logger.error(f"Unknown Core Exception {e}")
    #     exit("core error") if self.quiet_mode else exit()
    # except SEETMIOException as e:
    #     logger.error(f"Unknown IO Exception {e}")
    #     exit("io error") if self.quiet_mode else exit()
    # except SEETMException as e:
    #     logger.error(f"Unknown Base Exception {e}")
    #     exit("base error") if self.quiet_mode else exit()
    # # Default exceptions
    # except KeyboardInterrupt:
    #     logger.info(f"Gracefully terminating SEETM CLI Mapper...")
    #     exit("keyboard error") if self.quiet_mode else exit()
    # except MemoryError as e:
    #     logger.info(f"Out of memory exception occurred "
    #                 f"while loading the model. {e}")
    #     exit("memory error") if self.quiet_mode else exit()
    # except OSError as e:
    #     logger.info(f"OSError exception occurred "
    #                 f"while loading the model. {e}")
    #     exit("os error") if self.quiet_mode else exit()
    # except Exception as e:
    #     logger.error(f"Unknown CLI Exception. {e}")
    #     exit("generic error") if self.quiet_mode else exit()


if __name__ == "__main__":
    logger.error("This script cannot be directly executed. "
                 "please use the 'seetm' CLI tool instead.")
    exit(code=1)



import json
import logging
from typing import Dict, Text

from seetm.shared.constants import (
    InterfaceType,
    Encoding,
    FilePermission,
    Metrics,
    Config,
    DEFAULT_DATA_PATH,
    MappingMethod,
    DEFAULT_PORT,
    DEFAULT_HOST_LOCAL,
    ConfigType,
    DEFAULT_HOST_DEC,
)
from seetm.shared.exceptions.config import (
    InvalidInterfaceException,
    InvalidConfigKeyException,
)
from seetm.utils.io import (
    read_yaml_file,
    dir_exists,
)

logger = logging.getLogger(__name__)


def get_init_configs(
        config_path: Text = None,
        data_path: Text = None,
        instance: Text = None,
        method: Text = None,
        metric: Text = None,
        case: bool = True,
        level: Text = None,
        port: int = None,
        interface: Text = None,
) -> Dict:
    # setting default config file
    # path if not specified
    if not config_path:
        config_path = Config.DEFAULT_CONFIG_PATH

    # interface warnings
    if not interface:
        interface = InterfaceType.NONE

    if interface not in [InterfaceType.MAPPER, InterfaceType.SERVER, InterfaceType.EVALUATOR, InterfaceType.EXTRACTOR]:
        logger.error("An invalid interface has been specified. Valid "
                     "interfaces are MAPPER, SERVER, EVALUATOR, and EXTRACTOR")
        raise InvalidInterfaceException()

    # loading default configs
    try:
        yaml_content = read_yaml_file(
            yaml_file=config_path,
            encoding=Encoding.UTF8,
            mode=FilePermission.READ,
            version_check=False,
        )

        if not yaml_content:
            logger.warning(f"The given config file '{config_path}' is invalid. "
                           f"SEETM will utilize default configs instead")
    except FileNotFoundError:
        yaml_content = None
        logger.warning(f"Could not locate the '{config_path}' file. "
                       f"SEETM will utilize default configs instead")

    yaml_content_json = json.loads(json.dumps(yaml_content, ensure_ascii=False))
    default_configs = get_default_configs(section=ConfigType.ALL)

    if yaml_content_json:
        # validate keys
        invalid_keys = list()
        for key in yaml_content_json.keys():
            if key not in Config.VALID_MAIN_KEYS:
                invalid_keys.append(key)

        if Config.BASE_CONFIGS_KEY in yaml_content_json:
            base_dict = dict()
            for element in yaml_content_json[Config.BASE_CONFIGS_KEY]:
                base_dict = {**base_dict, **element}

            for key in base_dict.keys():
                if key not in Config.VALID_BASE_KEYS:
                    invalid_keys.append(key)

        if Config.SERVER_CONFIGS_KEY in yaml_content_json:
            server_dict = dict()
            for element in yaml_content_json[Config.SERVER_CONFIGS_KEY]:
                server_dict = {**server_dict, **element}

            for key in server_dict.keys():
                if key not in Config.VALID_SERVER_KEYS:
                    invalid_keys.append(key)

        if invalid_keys:
            logger.warning(f"Found Invalid Configuration YAML keys: {', '.join(invalid_keys)}")
            raise InvalidConfigKeyException()

        for main_key in yaml_content_json.keys():
            main_key_dict = dict()
            for element in yaml_content_json[main_key]:
                main_key_dict = {**main_key_dict, **element}

            for sub_key, value in main_key_dict.items():
                if sub_key == Config.DATA_PATH_KEY and not data_path and not instance:
                    if not dir_exists(value) and interface != InterfaceType.EVALUATOR:
                        logger.warning("Data path specified is invalid. SEETM will treat "
                                       f"'{default_configs[Config.BASE_CONFIGS_KEY][Config.DATA_PATH_KEY]}' "
                                       f"as the default data path.")
                        continue
                elif sub_key == Config.METHOD_KEY and not method:
                    if value not in [MappingMethod.IPA, MappingMethod.RULE_BASED] and \
                            interface != InterfaceType.EVALUATOR:
                        logger.warning("Mapping method specified is invalid. SEETM will utilize "
                                       f"'{default_configs[Config.BASE_CONFIGS_KEY][Config.METHOD_KEY]}' "
                                       f"as the default mapping method.")
                        continue
                elif sub_key == Config.METRIC_KEY and not metric:
                    if value not in [Metrics.EDIT_DISTANCE, Metrics.SEQUENCE_MATCHING] and \
                            interface == InterfaceType.EVALUATOR:
                        logger.warning("Metric specified is invalid. SEETM will utilize "
                                       f"'{default_configs[Config.BASE_CONFIGS_KEY][Config.METRIC_KEY]}' "
                                       f"as the default evaluation metric.")
                        continue
                elif sub_key == Config.PORT_KEY and not port:
                    if not isinstance(value, int) and interface == InterfaceType.SERVER:
                        logger.warning("Port specified is either invalid or unavailable. SEETM will utilize "
                                       f"'{default_configs[Config.SERVER_CONFIGS_KEY][Config.PORT_KEY]}' "
                                       f"as the default port to start SEETM server at.")
                        continue
                elif sub_key == Config.HOST_KEY:
                    if value not in [DEFAULT_HOST_LOCAL, DEFAULT_HOST_DEC] and \
                            interface == InterfaceType.SERVER:
                        logger.warning("Host specified is either invalid or unavailable. SEETM will utilize "
                                       f"'{default_configs[Config.SERVER_CONFIGS_KEY][Config.HOST_KEY]}' "
                                       f"as the default host to start SEETM server at.")
                        continue
                default_configs[main_key][sub_key] = value

    if data_path and interface == InterfaceType.EXTRACTOR:
        default_configs[Config.BASE_CONFIGS_KEY][Config.DATA_PATH_KEY] = data_path
        logger.warning("Data path specified in the config file will be ignored "
                       "since --data-path argument was set via the CLI")
    if instance and interface != InterfaceType.EVALUATOR:
        default_configs[Config.BASE_CONFIGS_KEY][Config.DATA_PATH_KEY] = [instance]
        logger.warning("Data path specified in the config file will be ignored "
                       "since --instance argument was set via the CLI")
    if method:
        if method in [MappingMethod.IPA, MappingMethod.RULE_BASED, MappingMethod.TRANSLATE]:
            default_configs[Config.BASE_CONFIGS_KEY][Config.METHOD_KEY] = method
            logger.warning("Mapping method specified in the config file will be ignored "
                           "since --method argument was set via the CLI")
    if port and interface == InterfaceType.SERVER:
        if isinstance(port, int):
            default_configs[Config.SERVER_CONFIGS_KEY][Config.PORT_KEY] = port
            logger.warning("Port specified in the config file will be ignored "
                           "since --port argument was set via the CLI")
    if isinstance(case, bool) and case is False:
        default_configs[Config.BASE_CONFIGS_KEY][Config.CASE_SENSITIVE_KEY] = case
        logger.warning("Case sensitivity specified in the config file will be ignored "
                       "since --no-case argument was set via the CLI")
    if interface == InterfaceType.EVALUATOR:
        if metric in [Metrics.EDIT_DISTANCE, Metrics.SEQUENCE_MATCHING]:
            default_configs[Config.BASE_CONFIGS_KEY][Config.METRIC_KEY] = metric
            logger.warning("Evaluation metric specified in the config file will be ignored "
                           "since --metric argument was set via the CLI")

    return default_configs


def get_default_configs(section: Text = ConfigType.ALL) -> Dict:
    default_configs = {
        "seetm_base_configs": {
            "data_path": DEFAULT_DATA_PATH,
            "method": MappingMethod.RULE_BASED,
            "metric": Metrics.EDIT_DISTANCE,
            "case_sensitive": True,
        },
        "seetm_server_configs": {
            "host": DEFAULT_HOST_LOCAL,
            "port": DEFAULT_PORT,
        }
    }

    if section == ConfigType.BASE:
        return default_configs[Config.BASE_CONFIGS_KEY]
    elif section == ConfigType.SERVER:
        return default_configs[Config.SERVER_CONFIGS_KEY]
    else:
        return default_configs



import json
import logging
import os
from typing import Dict, NoReturn, Tuple, Text

import ruamel.yaml as yml

from seetm.shared.constants import (
    Config,
    Validity,
    ServerConfigType,
    Encoding,
    FilePermission,
)
from seetm.shared.exceptions.config import (
    SEETMConfigException,
)
from seetm.shared.exceptions.io import (
    ConfigFileNotFoundException,
)
from seetm.shared.exceptions.server import (
    InvalidServerConfigsException,
    ServerConfigsPersistException,
    InvalidConfigurationTypeSpecifiedException,
    CustomConfigsNotFoundException,
)
from seetm.utils.io import (
    nlu_data_dir_exists,
)
from seetm.utils.io import read_yaml_file

logger = logging.getLogger(__name__)


class ServerConfigs:
    def __init__(self):
        self.configs_yml = None
        try:
            self._initialize()
        except Exception as e:
            logger.error(f"Exception occurred while initializing server configs. {e}")
            raise InvalidServerConfigsException(e)

    def _initialize(self) -> NoReturn:
        try:
            yaml_content = read_yaml_file(
                yaml_file=Config.DEFAULT_CONFIG_PATH,
                version_check=False,
            )
            self.configs_yml = yaml_content
            self.configs = json.loads(json.dumps(yaml_content, ensure_ascii=False))
            for key in self.configs.keys():
                key_content_dict = dict()
                for element_dict in self.configs[key]:
                    for element_key, element_value in element_dict.items():
                        key_content_dict[element_key] = element_value

                self.configs[key] = key_content_dict
        except FileNotFoundError:
            raise ConfigFileNotFoundException("Could not locate the 'seetm_config.yml' file. "
                                              "Make sure it is in the root directory")
        except Exception as e:
            logger.error(f"Exception occurred while retrieving 'seetm_config.yml'. {e}")
            raise SEETMConfigException(e)

    @staticmethod
    def validate(configs: Dict) -> Tuple[bool, Dict]:
        validate_status_obj = {
            "keys": Validity.NOTSET,
            Config.DATA_PATH_KEY: Validity.NOTSET,
        }
        validate_status = False

        try:
            keys = list(configs)
            if sorted(keys) != sorted([Config.DATA_PATH_KEY]):
                logger.error("Failed to validate updated server config keys")
                validate_status_obj['keys'] = Validity.INVALID
            else:
                validate_status_obj['keys'] = Validity.VALID

            # data path validation
            if not nlu_data_dir_exists(dir_path=configs[Config.DATA_PATH_KEY]):
                logger.error("Failed to validate data path in updated server configs")
                validate_status_obj[Config.DATA_PATH_KEY] = Validity.INVALID
            else:
                validate_status_obj[Config.DATA_PATH_KEY] = Validity.VALID

            # check if all keys are valid
            if Validity.INVALID in validate_status_obj.values():
                validate_status = False
            else:
                validate_status = True
            return validate_status, validate_status_obj
        except Exception as e:
            logger.error(f"Exception occurred while validating server configurations. {e}")
            return validate_status, validate_status_obj

    def update_and_persist(self, updated_configs: Dict, validate: bool = True):
        try:
            # latest configs
            self._initialize()
            
            if validate:
                if not self.validate(updated_configs)[0]:
                    raise InvalidServerConfigsException()

            seetm_base_configs = self.configs_yml.get(Config.BASE_CONFIGS_KEY)
            for k, v in updated_configs.items():
                for index, element in enumerate(seetm_base_configs):
                    if k in dict(element).keys():
                        seetm_base_configs[index] = {k: v}
            self.configs_yml[Config.BASE_CONFIGS_KEY] = seetm_base_configs

            # modifying the latest configs
            with open(
                    file=os.path.join(Config.DEFAULT_CONFIG_PATH),
                    encoding=Encoding.UTF8,
                    mode=FilePermission.WRITE
            ) as server_configs:
                yaml = yml.YAML()
                yaml.indent(sequence=4, offset=2)
                yaml.dump(self.configs_yml, server_configs)

            # retrieve saved configs
            self._initialize()

            logger.debug("Saved server configs")
        except Exception as e:
            raise ServerConfigsPersistException(e)

    def retrieve(self, config_type: Text = ServerConfigType.JSON, custom_configs: bool = False) -> Text:
        if not config_type or config_type not in [ServerConfigType.JSON, ServerConfigType.NONE]:
            raise InvalidConfigurationTypeSpecifiedException()

        if custom_configs:
            try:
                sinhala = bool(os.environ.get("SINHALA_ENABLED"))
                app_env = os.environ.get("APP_ENV")
                app_theme = os.environ.get("APP_THEME")
                self.configs["custom_configs"] = {
                    "sinhala_enabled": sinhala,
                    "app_theme": app_theme,
                    "app_env": app_env,
                }
            except Exception as e:
                logger.error("Failed to retrieve custom server configs.")
                raise CustomConfigsNotFoundException(e)

        if config_type == ServerConfigType.JSON:
            return json.dumps(self.configs, indent=4, ensure_ascii=False).encode(Encoding.UTF8).decode()
        else:
            return self.configs



import logging
from typing import NoReturn, Dict, Text

from seetm.core.token_mapper import TokenMapper
from seetm.shared.constants import (
    MappingMethod,
    Config,
    DEFAULT_NLU_EXAMPLES_TAG,
    DEFAULT_NLU_LOCATION_TAG,
)
from seetm.shared.exceptions.base import SEETMException
from seetm.shared.exceptions.core import (
    SEETMCoreException,
    MapperException,
    InvalidDataInstanceException,
    ExportableInitializationException,
)
from seetm.shared.exceptions.io import (
    SEETMIOException,
    EmptyNLUDatasetException,
    InvalidFileExtensionException,
    InvalidNLUDatasetException,
    NLUFileNotFoundException,
    YAMLFormatException,
)
from seetm.shared.exportable import Exportable
from seetm.utils.io import (
    dir_exists,
    get_rasa_testing_data,
)

logger = logging.getLogger(__name__)


class SEETMCLIMapper:
    def __init__(
            self,
            configs: Dict,
            persist: bool = False,
            quiet_mode: bool = False,
    ) -> NoReturn:
        self.data_path = configs[Config.BASE_CONFIGS_KEY][Config.DATA_PATH_KEY]
        self.method = configs[Config.BASE_CONFIGS_KEY][Config.METHOD_KEY]
        self.case_sensitive = configs[Config.BASE_CONFIGS_KEY][Config.CASE_SENSITIVE_KEY]
        self.persist = persist
        self.quiet_mode = quiet_mode
        self.token_mapper = TokenMapper(
            method=self.method,
            case_sensitive=self.case_sensitive,
            quiet_mode=self.quiet_mode,
        )

    def _map(self, instance: Text) -> Text:
        return self.token_mapper.map(data_instance=instance)

    def run(self) -> NoReturn:
        logger.debug("Initializing SEETM CLI Mapper...")

        # Deprecations, Errors and Warnings
        # for SEETM CLI arguments
        if self.method not in [MappingMethod.IPA, MappingMethod.RULE_BASED]:
            logger.error("Methods other than IPA and Rule-based are not yet supported by SEETM.")
            exit()

        if isinstance(self.data_path, str):
            if not dir_exists(self.data_path):
                raise InvalidNLUDatasetException()

            nlu_dataset = get_rasa_testing_data(
                testing_data_dir=self.data_path,
                preserve_location=True,
            )

            if not nlu_dataset:
                EmptyNLUDatasetException()

            mapped_list = list()
            for intent in nlu_dataset.keys():
                nlu_file = nlu_dataset[intent][DEFAULT_NLU_LOCATION_TAG]
                for example in nlu_dataset[intent][DEFAULT_NLU_EXAMPLES_TAG]:
                    mapped_instance = self._map(instance=example)
                    mapped_list.append({
                        "instance": example,
                        "mapped_instance": mapped_instance,
                        "intent": intent,
                        "nlu_file": nlu_file,
                        "method": self.method,
                    })

            exportable = Exportable(exportable=mapped_list)
            if not self.quiet_mode:
                exportable.inspect()
            if self.persist:
                exportable.persist()

        else:
            instance = self.data_path[0]
            mapped_instance = self._map(instance=instance)
            mapped_list = [
                {
                    "instance": instance,
                    "mapped_instance": mapped_instance,
                    "intent": "none",
                    "nlu_file": "none",
                    "method": self.method,
                }
            ]

            exportable = Exportable(exportable=mapped_list)
            if not self.quiet_mode:
                exportable.inspect()
            if self.persist:
                exportable.persist()



from seetm.shared.exceptions.base import SEETMException


class SEETMCoreException(SEETMException):
    pass


class MapperException(SEETMCoreException):
    pass


class InvalidDataInstanceException(SEETMCoreException):
    pass


class InvalidNumberOfTokensException(SEETMCoreException):
    pass


class MapLoadingException(SEETMCoreException):
    pass


class ExportableInitializationException(SEETMCoreException):
    pass


class ExportablePersistException(SEETMCoreException):
    pass


class EvaluationPersistException(SEETMCoreException):
    pass


class EvaluationDatasetExportException(SEETMCoreException):
    pass


class MappingExportException(SEETMCoreException):
    pass


class InvalidEvaluationLevelException(SEETMCoreException):
    pass


class InvalidEvaluationMethodException(SEETMCoreException):
    pass


class InvalidMappingMethodException(SEETMCoreException):
    pass


class Word2VecModelTrainingException(SEETMCoreException):
    pass


class Word2VecModelNotFoundException(SEETMCoreException):
    pass


class ZeroMappedTokensException(SEETMCoreException):
    pass


class ZeroExtractableTokensException(SEETMCoreException):
    pass


class ZeroValidTokensException(SEETMCoreException):
    pass


class OutDatedEvaluationDatasetException(SEETMCoreException):
    pass


class ExtractorCleanupException(SEETMCoreException):
    pass



import logging
import os
import re
import shutil
import sys
from collections import OrderedDict as OrderedDictColl
from datetime import datetime
from os import path
from typing import (
    List,
    Optional,
    Text,
    OrderedDict,
    Dict,
    NoReturn,
    Union,
    Tuple,
)
from uuid import uuid4

from rasa.shared.data import (
    get_data_files,
    is_nlu_file,
)
from ruamel import yaml as yaml
from ruamel.yaml.comments import CommentedMap
from ruamel.yaml.error import YAMLError

from seetm.shared.constants import (
    Encoding,
    FilePermission,
    TermColor,
    DEFAULT_DATA_PATH,
    DEFAULT_NLU_YAML_TAG,
    DEFAULT_VERSION_YAML_TAG,
    DEFAULT_YAML_VERSION,
    YAML_EXTENSIONS,
    DEFAULT_NLU_INTENT_TAG,
    DEFAULT_NLU_EXAMPLES_TAG,
    DEFAULT_NLU_LOCATION_TAG,
    DEFAULT_INIT_DEST_DIR_NAME,
    DEFAULT_EVALUATION_DATA_PATH,
    DEFAULT_EVAL_YAML_TAG,
    DEFAULT_EVAL_TOKEN_TAG,
    DEFAULT_EVAL_TEST_TOKEN_TAG,
)
from seetm.shared.exceptions.io import (
    YAMLFormatException,
    NLUFileNotFoundException,
    InvalidFileExtensionException,
    FileSizeInspectingException,
    EvalFileNotFoundException, YAMLFileWriteException,
)

logger = logging.getLogger(__name__)


def get_seetm_init_caches(
        dir_path: Text = DEFAULT_INIT_DEST_DIR_NAME,
) -> Optional[List]:
    """
    Returns all dime cache directories in a
    specified directory path

    Args:
        dir_path: path to retrieve the list of
            initial DIME cache directories and
            files

    Returns:
        list of initial DIME caching directories
        and files
    """
    return [cache_dir for cache_dir in os.listdir(dir_path)
            if str(cache_dir).startswith(".dime_init_")]


def _get_dir_file_list(
        dir_path: Text = DEFAULT_DATA_PATH,
        file_suffixes: List = YAML_EXTENSIONS
) -> Optional[List]:
    """
    Returns the list files available in a given
    directory path and its subdirectories. able
    to check extensions if the extensions are
    provided as a list

    Args:
        dir_path: path where to read the list of directories
        file_suffixes: list of file extensions to look for. if
            not provided, the method will output all the files
            available in the given destination directory

    Returns:
        list of files available, or None
    """

    file_paths = list()

    for (dir_path, dir_names, file_names) in os.walk(dir_path):
        if file_suffixes:
            file_paths += [path.join(dir_path, file) for file in file_names if str(file).endswith(tuple(file_suffixes))]
        else:
            file_paths += [path.join(dir_path, file) for file in file_names]
    return file_paths


def get_all_existing_file_list(
        dir_path: Text = DEFAULT_DATA_PATH,
) -> Optional[List]:
    """
    Returns all files and directories existing
    in the specified destination directory and
    subdirectories

    Args:
        dir_path: destination directory path where
            to get the list of files and directories

    Returns:
        list of files and directories available, or
            None
    """

    file_list_all = list()

    for (dir_path, dir_names, file_names) in os.walk(dir_path):
        file_list_all += [dir_ for dir_ in dir_names]
        file_list_all += [file for file in file_names]

    return file_list_all


def get_existing_toplevel_file_list(
        dir_path: Text = DEFAULT_DATA_PATH,
        exclude: List = None
) -> Optional[List]:
    """
    Returns the list files available in a given
    directory path but not subdirectories. able
    to ignore files if the provided as a list under
    exclude argument

    Args:
        dir_path: path where to read the list of directories
        exclude: list of file names to ignore

    Returns:
        list of files available except the files mentioned in
            exclude list, or None
    """

    files = os.listdir(dir_path)
    if exclude:
        files = list(set(files).difference(set(exclude)))
    return files


def get_timestamp_str(sep: Text = "-", uuid: bool = False) -> Text:
    """
    Generates a timestamped string and attaches a unique
    ID using UUID if specified and returns it

    Args:
        sep: seperator to separate date, time and uuid
        uuid: if True, a unique UUID will be attached
            at the end of the timestamped string
            generated

    Returns:
        timestamped string
    """
    if uuid:
        return datetime.now().strftime('%Y%m%d' + sep + '%H%M%S' + sep) \
               + str(uuid4())
    else:
        return datetime.now().strftime('%Y%m%d' + sep + '%H%M%S')


def dir_exists(dir_path: Text = None) -> bool:
    """
    Checks if the specified directory path exists

    Args:
        dir_path: path of the directory to check
            the existence

    Returns:
        True if path exists, else False
    """
    return path.exists(dir_path) and path.isdir(dir_path)


def file_exists(file_path: Text = None) -> bool:
    """
    Checks if the specified file path exists

    Args:
        file_path: path of the file to check
            the existence

    Returns:
        True if path exists, else False
    """
    return path.exists(file_path) and path.isfile(file_path)


def file_size(
        file_path: Union[Text, List],
        reverse: bool = False
) -> Union[Dict, Tuple[float, Text]]:
    try:
        if isinstance(file_path, Text):
            size_in_bytes = os.path.getsize(file_path)
            if size_in_bytes >= 1000000000:
                size = size_in_bytes / 1000000000
                units = "GB"
            elif size_in_bytes >= 1000000:
                size = size_in_bytes / 1000000
                units = "MB"
            elif size_in_bytes >= 1000:
                size = size_in_bytes / 1000
                units = "KB"
            else:
                size = size_in_bytes
                units = "Bytes"
            return size, units
        else:
            if reverse:
                file_path.reverse()
            file_sizes_dict = {file: file_size(file_path=file) for file in file_path}
            return file_sizes_dict
    except Exception as e:
        raise FileSizeInspectingException(e)


def read_yaml_file(
        yaml_file: Text = None,
        encoding: Text = Encoding.UTF8,
        mode: Text = FilePermission.READ,
        yaml_version: Text = DEFAULT_YAML_VERSION,
        version_check: bool = True,
) -> Optional[OrderedDict]:
    """
    Reads the content of a specified YAML file.

    Args:
        yaml_file: YAML file to read the content from
        encoding: encoding of the specified file
        mode: read mode
        yaml_version: version of the YAML file, 2.0 is
            preferred
        version_check: True if required to check the
            version, else False

    Returns:
        YAML file content as an ordered dictionary, or
            None
    """
    if not yaml_file:
        return OrderedDictColl()

    with open(file=yaml_file, encoding=encoding, mode=mode) as file_stream:
        try:
            yaml_content = yaml.round_trip_load(file_stream, preserve_quotes=False)
        except YAMLError as e:
            raise YAMLFormatException(e)

    if not yaml_content:
        return None

    if version_check:
        if DEFAULT_VERSION_YAML_TAG not in yaml_content:
            logger.warning("The YAML file is not properly versioned. "
                           "the expected version is 2.0")
        else:
            if yaml_content[DEFAULT_VERSION_YAML_TAG] != yaml_version:
                logger.warning("The YAML file does not contain the expected "
                               "rasa YAML version. the expected version is "
                               "2.0")

    return yaml_content


def _find_yaml_collection(
        yaml_content: OrderedDict = None,
        yaml_collection_tag: Text = DEFAULT_NLU_YAML_TAG,
) -> OrderedDict:
    """
    Given the content of a YAML file, finds and returns
    the specified YAML collection.

    Args:
        yaml_content: content of a YAML file as an ordered
            dictionary
        yaml_collection_tag: tag of the collection to be
            read

    Returns:
        The YAML collection extracted out of the YAML content
            as an ordered dictionary
    """
    if yaml_content is None:
        return OrderedDictColl()
    if yaml_collection_tag not in yaml_content:
        return OrderedDictColl()
    return yaml_content[yaml_collection_tag]


def _sanitize_rasa_nlu_data(
        collection_content: Optional[OrderedDict],
        case_sensitive: bool = True
) -> Optional[Dict]:
    """
    Receives a collection of already retrieved
    RASA NLU examples and cleans it. Returns a
    sanitized non-empty collection of nlu example
    sentences.

    Args:
        collection_content: input data as an ordered
            dictionary
        case_sensitive:  if True, reads all data as
            lowercase, or else preserves the case and
            reads the files as is

    Returns:
        sanitized NLU data as a dictionary which
            has the list of examples under 'examples'
            under the key 'intent'
    """

    if not collection_content:
        return None

    nlu_examples = {
        intent[DEFAULT_NLU_INTENT_TAG]: [
            str.strip(instance) for instance in
            (str.split(
                re.sub("- ", "", intent[DEFAULT_NLU_EXAMPLES_TAG]),
                "\n"
            ))
            if str.strip(instance) not in ['']
        ]
        for intent in collection_content
        if len(intent) == 2
    }

    if not case_sensitive:
        nlu_examples = {
            intent: [str.lower(example) for example in examples]
            for intent, examples
            in nlu_examples.items()
        }
    return nlu_examples


def get_unique_list(list_of_data: Optional[List] = None) -> Optional[List]:
    """
    Converts a given list to a list that only contains
    unique elements. Eliminates duplicate elements by
    converting to a set and back to a list.

    Args:
        list_of_data: list of elements

    Returns:
        list containing only unique elements, or None
    """

    if not list_of_data:
        return None
    return list(set(list_of_data))


def get_rasa_testing_data(
        testing_data_dir: Text = DEFAULT_DATA_PATH,
        file_ext: List = YAML_EXTENSIONS,
        case_sensitive: bool = True,
        preserve_location: bool = False
) -> Optional[Dict]:
    testing_data = dict()
    logger.debug("Initializing testing data...")

    if file_exists(file_path=testing_data_dir):
        file_extension = os.path.splitext(testing_data_dir)[-1]
        if file_extension in YAML_EXTENSIONS:
            testing_data_files = [testing_data_dir]
        else:
            raise InvalidFileExtensionException("The specified file is not a valid "
                                                "YAML file")
    else:
        testing_data_files = _get_dir_file_list(dir_path=testing_data_dir, file_suffixes=file_ext)
    logger.info(f"{len(testing_data_files)} YAML files were found in the NLU data directory.")

    if testing_data_files:
        logger.debug(f"List of files found: {', '.join([os.path.split(fp)[-1] for fp in testing_data_files])}")

    for file in testing_data_files:
        try:
            nlu_data = read_yaml_file(
                yaml_file=file,
                version_check=True
            )
        except FileNotFoundError:
            raise NLUFileNotFoundException("Could not locate one or more "
                                           "NLU testing data YAML files")
        nlu_content = _find_yaml_collection(
            yaml_content=nlu_data,
            yaml_collection_tag=DEFAULT_NLU_YAML_TAG
        )
        cleaned_nlu_content = _sanitize_rasa_nlu_data(
            collection_content=nlu_content,
            case_sensitive=case_sensitive
        )

        if not cleaned_nlu_content:
            continue

        for intent, examples in cleaned_nlu_content.items():
            if intent in testing_data:
                testing_data[intent][DEFAULT_NLU_EXAMPLES_TAG] += examples
                if file not in testing_data[intent][DEFAULT_NLU_LOCATION_TAG] and preserve_location:
                    testing_data[intent][DEFAULT_NLU_LOCATION_TAG] += file

            else:
                testing_data[intent] = {DEFAULT_NLU_EXAMPLES_TAG: [], DEFAULT_NLU_LOCATION_TAG: []}
                testing_data[intent][DEFAULT_NLU_EXAMPLES_TAG] = examples
                if preserve_location:
                    testing_data[intent][DEFAULT_NLU_LOCATION_TAG] = [os.path.split(file)[-1]]

    return testing_data


def get_seetm_eval_data(
        eval_data_dir: Text = DEFAULT_EVALUATION_DATA_PATH,
        file_ext: List = YAML_EXTENSIONS,
        case_sensitive: bool = True,
        preserve_location: bool = False
) -> Optional[List]:
    logger.debug("Initializing evaluation data...")
    eval_dataset = list()

    if file_exists(file_path=eval_data_dir):
        file_extension = os.path.splitext(eval_data_dir)[-1]
        if file_extension in YAML_EXTENSIONS:
            eval_data_files = [eval_data_dir]
        else:
            raise InvalidFileExtensionException("The specified file is not a valid "
                                                "YAML file")
    else:
        eval_data_files = _get_dir_file_list(dir_path=eval_data_dir, file_suffixes=file_ext)
    logger.info(f"{len(eval_data_files)} YAML files were found in the Evaluation data directory.")

    if eval_data_files:
        logger.debug(f"List of files found: "
                     f"{', '.join([os.path.split(fp)[-1] for fp in eval_data_files])}")

    for file in eval_data_files:
        try:
            eval_data = read_yaml_file(
                yaml_file=file,
                version_check=False
            )
        except FileNotFoundError:
            raise EvalFileNotFoundException("Could not locate one or more "
                                            "eval YAML files")
        eval_content = _find_yaml_collection(
            yaml_content=eval_data,
            yaml_collection_tag=DEFAULT_EVAL_YAML_TAG
        )

        if not eval_content:
            continue

        added_tokens = list()
        for sentence_pair in eval_content:
            token = sentence_pair.get(DEFAULT_EVAL_TOKEN_TAG)
            test_token = sentence_pair.get(DEFAULT_EVAL_TEST_TOKEN_TAG)

            if not case_sensitive:
                token = str.lower(token)
                test_token = str.lower(test_token)

            sentence_pair_dict = {
                "token": token,
                "test_token": test_token,
                "intent": None
            }

            if preserve_location:
                sentence_pair_dict.update({"location": [os.path.split(file)[-1]]})

            if sentence_pair_dict['token'] not in added_tokens:
                eval_dataset.append(sentence_pair_dict)
                added_tokens.append(sentence_pair_dict['token'])
            else:
                logger.warning(f"Found duplicate mappings for {sentence_pair_dict['token']}. "
                               f"Only one mapping will be evaluated")

    return eval_dataset


def set_cli_color(text_content: Text = None, color: str = TermColor.NONE_C):
    """
    wraps inbound string instances with ASCII
    color codes and returns a colored string
    that can be printed on the terminals

    Args:
        text_content: string instance to wrap
        color: color code to be visible on
            terminals

    Returns:
        wrapped string with the specified color
            that is ready to be printed on a
            terminal
    """
    return color + str(text_content) + TermColor.END_C


def update_sys_path(path_to_add: Text) -> NoReturn:
    """
    Appends a given path to the list of system
    paths. Can utilize this method to resolve
    import conflicts if packages or modules are
    being ignored at runtime in a specific
    location

    Args:
        path_to_add: path of the directory that
            that should be added to the system
            path list

    Returns:
        no return
    """
    sys.path.insert(0, path_to_add)


def series_to_json_serializable(series: Union[Dict, List]):
    if isinstance(series, dict):
        return {k: float(v) for k, v in series.items()}
    elif isinstance(series, list):
        return [float(v) for v in series]


def nlu_data_dir_exists(dir_path: Text) -> bool:
    if not os.path.exists(dir_path):
        return False

    data_list = get_data_files(paths=[dir_path], filter_predicate=is_nlu_file)
    if len(data_list) == 0:
        return False

    return True


def write_yaml_file(
        file_path: Text = None,
        content: CommentedMap = None,
        encoding: Text = Encoding.UTF8,
        mode: Text = FilePermission.WRITE,
        preserve_quotes: bool = True,
) -> NoReturn:
    try:
        if not file_path or not content:
            raise Exception()

        with open(file_path, encoding=encoding, mode=mode) as yaml_file:
            yml = yaml.YAML()
            yml.indent(sequence=4, offset=2)
            yml.preserve_quotes = preserve_quotes
            yaml.round_trip_dump(content, yaml_file)

    except Exception as e:
        raise YAMLFileWriteException(e)


def delete_file(
        file_path: Union[Text, List]
) -> NoReturn:
    if isinstance(file_path, str):
        if file_exists(file_path):
            os.remove(path=file_path)
    else:
        for file in file_path:
            delete_file(file_path=file)


def delete_dir(
        dir_path: Union[Text, List],
        force: bool = False
) -> NoReturn:
    if isinstance(dir_path, str):
        if dir_exists(dir_path=dir_path):
            if force:
                shutil.rmtree(path=dir_path)
            else:
                os.rmdir(path=dir_path)
    else:
        for dir_ in dir_path:
            delete_dir(dir_path=dir_)


def delete_top_level_files(
        dir_path: Union[Text, List],
        ext: List = None
) -> NoReturn:
    if isinstance(dir_path, str):
        all_files = get_existing_toplevel_file_list(dir_path=dir_path)
        filtered_files = list()
        for file in all_files:
            for extension in ext:
                if str.endswith(file, extension):
                    filtered_files.append(os.path.join(dir_path, file))
                else:
                    continue
        if filtered_files:
            delete_file(file_path=filtered_files)
    else:
        for dir_ in dir_path:
            delete_top_level_files(dir_path=dir_)



import logging
import os
import shutil
from typing import Text, NoReturn

import pkg_resources

from seetm.shared.constants import (
    DEFAULT_INIT_SRC_DIR_NAME,
    DEFAULT_INIT_DEST_DIR_NAME,
    DEFAULT_INIT_CACHE_DIR_NAME,
    DEFAULT_INIT_FILES_TO_EXCLUDE,
    ALLOWED_INIT_DIR_NAMES,
    DestinationDirType,
    RASA_DIRS_IN_SEETM_INIT,
)
from seetm.shared.exceptions.io import (
    InvalidInitDirException,
    ProjectExistsException,
)
from seetm.utils.io import (
    get_timestamp_str,
    get_seetm_init_caches,
    get_existing_toplevel_file_list,
)

logger = logging.getLogger(__name__)


class SEETMInit:
    def __init__(self, src: Text = DEFAULT_INIT_SRC_DIR_NAME):
        self._src = src
        self.src_path = None
        self.cache_path_timestamped = None
        self.cache_path_timestamped_path = None

    def build_scaffold(
            self,
            dest_path: Text = DEFAULT_INIT_DEST_DIR_NAME
    ) -> NoReturn:

        dest_path = str(dest_path).strip()
        if not dest_path or str(dest_path).strip() in ALLOWED_INIT_DIR_NAMES:
            dest_path = DEFAULT_INIT_DEST_DIR_NAME

        try:
            self.src_path = pkg_resources.resource_filename(__name__, self._src)
            self.cache_path_timestamped = DEFAULT_INIT_CACHE_DIR_NAME + get_timestamp_str()
            self.cache_path_timestamped_path = os.path.join(dest_path, self.cache_path_timestamped) + "\\"

            # removing existing caches
            cached_init_list = get_seetm_init_caches(dest_path)
            for elem in cached_init_list:
                shutil.rmtree(elem)

            dest_dir_type = self._verify_destination_dir(dest_path)
            if dest_dir_type == DestinationDirType.RASA:
                logger.info("The initialization directory is already a rasa directory. "
                            "only SEETM files will be created.")
                shutil.copytree(
                    src=self.src_path,
                    dst=self.cache_path_timestamped_path,
                    ignore=shutil.ignore_patterns('__pycache__', '__main__.py', 'data'),
                )
            elif dest_dir_type == DestinationDirType.VALID or dest_dir_type == DestinationDirType.EMPTY:
                shutil.copytree(
                    src=self.src_path,
                    dst=self.cache_path_timestamped_path,
                    ignore=shutil.ignore_patterns('__pycache__', '__main__.py'),
                )
            elif dest_dir_type == DestinationDirType.SEETM:
                raise ProjectExistsException("The initialization directory already contains a SEETM project. "
                                             "Initialization aborted.")
            else:
                raise InvalidInitDirException("The initialization directory seems invalid. "
                                              "Initialization aborted.")

            dir_contents = get_existing_toplevel_file_list(self.cache_path_timestamped_path)
            for elem in dir_contents:
                if elem in DEFAULT_INIT_FILES_TO_EXCLUDE:
                    continue

                abs_src_path = os.path.join(self.cache_path_timestamped_path, elem)
                rel_dst_path = os.path.join(dest_path, elem)
                shutil.move(src=abs_src_path, dst=rel_dst_path)
            shutil.rmtree(path=self.cache_path_timestamped_path)
            os.remove(path=os.path.join("./", "__init__.py"))

            logger.info("Initialized a new SEETM project.")
        except Exception as e:
            logger.error(f"Initializing SEETM project failed. more info: {e}")
            exit()

    def _verify_destination_dir(self, dest_path: Text = None) -> Text:
        if not dest_path:
            return DestinationDirType.INVALID

        existing_src_files = get_existing_toplevel_file_list(
            self.src_path,
            exclude=DEFAULT_INIT_FILES_TO_EXCLUDE
        )
        existing_dest_files = get_existing_toplevel_file_list(
            dir_path=dest_path,
            exclude=DEFAULT_INIT_FILES_TO_EXCLUDE
        )

        if not existing_dest_files:
            return DestinationDirType.EMPTY
        elif existing_dest_files:
            rasa_overlap = list(set(RASA_DIRS_IN_SEETM_INIT).intersection(set(existing_dest_files)))
            seetm_overlap = list(set(list(set(existing_src_files).intersection(set(existing_dest_files))))
                                 .difference(set(RASA_DIRS_IN_SEETM_INIT)))

            if seetm_overlap:
                logger.warning(f"Found overlapping SEETM files and/or directories: {', '.join(seetm_overlap)}")
                return DestinationDirType.SEETM
            elif rasa_overlap:
                logger.warning(f"Found overlapping RASA files and/or directories: {', '.join(rasa_overlap)}")
                return DestinationDirType.RASA
        return DestinationDirType.VALID



from seetm.shared.exceptions.base import SEETMException


class SEETMIOException(SEETMException):
    pass


class InvalidInitDirException(SEETMIOException):
    pass


class ProjectExistsException(SEETMIOException):
    pass


class YAMLFormatException(SEETMIOException):
    pass


class YAMLFileWriteException(SEETMIOException):
    pass


class NLUFileNotFoundException(SEETMIOException):
    pass


class EvalFileNotFoundException(SEETMIOException):
    pass


class InvalidFileExtensionException(SEETMIOException):
    pass


class InvalidNLUDatasetException(SEETMIOException):
    pass


class EmptyNLUDatasetException(SEETMIOException):
    pass


class InvalidPathSpecifiedException(SEETMIOException):
    pass


class ModelNotFoundException(SEETMIOException):
    pass


class ModelLoadException(SEETMIOException):
    pass


class FileSizeInspectingException(SEETMIOException):
    pass


class InvalidEvalDatasetException(SEETMIOException):
    pass


class EmptyEvalDatasetException(SEETMIOException):
    pass


class ConfigFileNotFoundException(SEETMIOException):
    pass



import json
import logging
import os.path
from typing import NoReturn, Dict, List, Tuple

import eng_to_ipa as ipa
import regex
from rich.progress import track
from ruamel.yaml.comments import (
    CommentedMap as OrderedDict,
    CommentedSeq as OrderedList,
)

from seetm.shared.constants import (
    Config,
    Encoding,
    FilePermission,
    DEFAULT_NLU_EXAMPLES_TAG,
    ENGLISH_TOKEN_PATTERN,
    DEFAULT_EVALUATION_DATA_PATH,
    EN_TO_PHONETICS_MAP_PATH,
    TOKEN_TO_TOKEN_MAP_PATH, YAML_EXTENSIONS,
)
from seetm.shared.exceptions.core import (
    EvaluationDatasetExportException,
    MappingExportException, ExtractorCleanupException,
)
from seetm.shared.exceptions.io import (
    EmptyNLUDatasetException,
    InvalidNLUDatasetException,
)
from seetm.utils.io import (
    dir_exists,
    get_rasa_testing_data,
    write_yaml_file,
    file_exists,
    delete_top_level_files,
    delete_file,
    read_yaml_file,
)
from seetm.utils.text_preprocessing import tokenize

logger = logging.getLogger(__name__)
EN_PATTERN = regex.compile(ENGLISH_TOKEN_PATTERN)


class SEETMExtractor:
    def __init__(
            self,
            configs: Dict,
            persist: bool = False,
            clean: bool = False,
            quiet_mode: bool = False,
    ) -> NoReturn:
        self.data_path = configs[Config.BASE_CONFIGS_KEY][Config.DATA_PATH_KEY]
        self.persist = persist
        self.clean = clean
        self.quiet_mode = quiet_mode
        self.progress_description_width = 30

    def extract(self) -> Tuple[List, List]:
        logger.debug("Initializing SEETM Extractor...")

        # loading the NLU dataset
        if not dir_exists(self.data_path):
            raise InvalidNLUDatasetException()

        nlu_dataset = get_rasa_testing_data(
            testing_data_dir=self.data_path,
            preserve_location=True,
        )

        if not nlu_dataset:
            EmptyNLUDatasetException()

        nlu_intents = nlu_dataset.keys() if self.quiet_mode else \
            track(
                nlu_dataset.keys(),
                description=f"{'Loading NLU dataset:':{self.progress_description_width}}"
            )

        nlu_example_list = list()
        for intent in nlu_intents:
            for example in nlu_dataset[intent][DEFAULT_NLU_EXAMPLES_TAG]:
                nlu_example_list.append(example)

        # cleaning existing files if requested
        if self.clean:
            try:
                delete_file(file_path=[EN_TO_PHONETICS_MAP_PATH, TOKEN_TO_TOKEN_MAP_PATH])
                delete_top_level_files(dir_path=DEFAULT_EVALUATION_DATA_PATH, ext=YAML_EXTENSIONS)
            except Exception as e:
                raise ExtractorCleanupException(e)

        mappable_token_list = list()
        unmappable_token_list = list()
        if not self.quiet_mode:
            nlu_example_list = track(
                nlu_example_list,
                description=f"{'Extracting tokens:':{self.progress_description_width}}"
            )
        for example in nlu_example_list:
            tokens = tokenize(instance=example)
            tokens = [str.lower(t) for t in tokens]
            for token in tokens:
                if len(EN_PATTERN.findall(token)) == 1:
                    if ipa.isin_cmu(token):
                        mappable_token_list.append(token) if token not in mappable_token_list else None
                    else:
                        unmappable_token_list.append(token) if token not in unmappable_token_list else None

        if self.persist:
            self._persist_evaluation_dataset(token_list=mappable_token_list)
            self._persist_en_to_phonetics_map(token_list=unmappable_token_list)
            self._persist_token_to_token_map(token_list=mappable_token_list + unmappable_token_list)
        else:
            if not self.quiet_mode:
                logger.warning("Evaluation dataset and maps will not be "
                               "persisted as `persist` was set to false`")

        return mappable_token_list, unmappable_token_list

    def _persist_evaluation_dataset(self, token_list: List) -> NoReturn:
        try:
            if not token_list:
                logger.warning("Not enough tokens were discovered "
                               "to generate the evaluation dataset")
                return

            # generate map skeletons
            number_of_tokens = len(token_list)
            try:
                existing_dataset_yml = read_yaml_file(
                    yaml_file=os.path.join(DEFAULT_EVALUATION_DATA_PATH, "token_eval.yml"),
                    encoding=Encoding.UTF8,
                    mode=FilePermission.READ,
                    version_check=False,
                )
            except OSError:
                existing_dataset_yml = {}
            existing_dataset = json.loads(json.dumps(existing_dataset_yml, ensure_ascii=False))
            existing_tokens = {
                mapping["token"]: mapping["test_token"] for mapping in existing_dataset["eval"]
                if mapping["token"] not in existing_dataset
            } if existing_dataset else dict()

            token_eval_dataset = OrderedDict({"eval": OrderedList()})
            if not self.quiet_mode:
                token_list = track(
                    token_list,
                    description=f"{'Generating Evaluation Dataset:':{self.progress_description_width}}"
                )
            for token in token_list:
                test_token = existing_tokens[token] if token in existing_tokens.keys() else ""
                token_eval_dataset.get("eval").append(
                    OrderedDict({
                        "token": token,
                        "test_token": test_token,
                    })
                )

            # inserting line breaks
            for key in range(1, len(token_eval_dataset.get("eval"))):
                token_eval_dataset["eval"].yaml_set_comment_before_after_key(
                    key=key,
                    before='\n'
                )

            write_yaml_file(
                file_path=os.path.join(DEFAULT_EVALUATION_DATA_PATH, "token_eval.yml"),
                content=token_eval_dataset,
                encoding=Encoding.UTF8,
                mode=FilePermission.WRITE,
                preserve_quotes=True,
            )
            if not self.quiet_mode:
                logger.info("Evaluation dataset skeleton was updated. Make sure to "
                            "manually add test_token mappings before evaluating")
                logger.info(f"Number of evaluation tokens persisted: {number_of_tokens}")
        except Exception as e:
            raise EvaluationDatasetExportException(e)

    def _persist_en_to_phonetics_map(self, token_list: List) -> NoReturn:
        try:
            if not token_list:
                logger.warning("Not enough tokens were discovered "
                               "to update IPA mappings")
                return

            # load existing map
            if file_exists(EN_TO_PHONETICS_MAP_PATH):
                with open(EN_TO_PHONETICS_MAP_PATH, encoding=Encoding.UTF8, mode=FilePermission.READ) as map_content:
                    existing_map = json.load(map_content)
            else:
                existing_map = {}
            existing_map = existing_map or {}
            updated_tokens = list()

            if not self.quiet_mode:
                token_list = track(
                    token_list,
                    description=f"{'Updating Map [Phonetics]:':{self.progress_description_width}}"
                )
            for token in token_list:
                if token not in existing_map.keys():
                    existing_map.update({token: ""})
                    updated_tokens.append(token)

            with open(EN_TO_PHONETICS_MAP_PATH, encoding=Encoding.UTF8, mode=FilePermission.WRITE) as updated_map:
                json.dump(existing_map, updated_map, ensure_ascii=False, indent=4)

            if not self.quiet_mode:
                logger.info("en_to_phonetics map was updated. Make sure to "
                            "manually add missing mappings before IPA mapping tasks")
                logger.info(f"Number of tokens updated: {len(updated_tokens)}")
        except Exception as e:
            raise MappingExportException(e)

    def _persist_token_to_token_map(self, token_list: List) -> NoReturn:
        try:
            if not token_list:
                logger.warning("Not enough tokens were discovered "
                               "to update Rule-based mappings")
                return

            # load existing map
            if file_exists(TOKEN_TO_TOKEN_MAP_PATH):
                with open(TOKEN_TO_TOKEN_MAP_PATH, encoding=Encoding.UTF8, mode=FilePermission.READ) as map_content:
                    existing_map = json.load(map_content)
            else:
                existing_map = {}
            existing_map = existing_map or {}
            updated_tokens = list()

            if not self.quiet_mode:
                token_list = track(
                    token_list,
                    description=f"{'Updating Map [Rule-based]:':{self.progress_description_width}}"
                )
            existing_keys = existing_map.keys() or list()
            existing_values = list()

            for key, values in existing_map.items():
                existing_values += values

            for token in token_list:
                if token in existing_keys or token in existing_values:
                    continue
                else:
                    existing_map.update({token: [token]})
                    updated_tokens.append(token)

            with open(TOKEN_TO_TOKEN_MAP_PATH, encoding=Encoding.UTF8, mode=FilePermission.WRITE) as updated_map:
                json.dump(existing_map, updated_map, ensure_ascii=False, indent=4)

            if not self.quiet_mode:
                logger.info("token_to_token map was updated. Make sure to manually "
                            "add missing mappings before Rule-based mapping tasks")
                logger.info(f"Number of tokens updated: {len(updated_tokens)}")
        except Exception as e:
            raise MappingExportException(e)



import logging
import re
from copy import copy
from logging import Formatter
from typing import Text

from seetm.shared.constants import (
    TermColor,
)

MAPPING = {
    'DEBUG': 35,  # white
    'INFO': 34,  # cyan
    'WARNING': 33,  # yellow
    'ERROR': 38,  # red
    'CRITICAL': 31,  # white on red bg
    'RESET': 12,
    'GRAY': 32,
}

PREFIX = '\033['
SUFFIX = '\033[0m'

logger = logging.getLogger(__name__)


class SEETMLoggingFormatter(Formatter):
    """
    A custom logging formatter for SEETM
    """

    def __init__(self, format_str):
        Formatter.__init__(self, format_str)

    def format(self, record) -> Text:
        """
        Formats a given logging record by
        wrapping it with required color codes

        Args:
            record: logging record to be wrapped

        Returns:
            formatted logging record
        """
        record_cpy = copy(record)
        try:
            formatted_record = Formatter.format(self, record_cpy)
            level_name = record_cpy.levelname
            colored_level_name = TermColor.LIGHTGREEN + str(level_name) + TermColor.END_C

            message = " - " + record_cpy.message
            asc_time = record_cpy.asctime
            name = record_cpy.name

            colored_message = TermColor.WHITE + str(message) + TermColor.END_C
            colored_asc_time = TermColor.WHITE + str(asc_time) + TermColor.END_C
            colored_name = TermColor.CYAN + str(name) + TermColor.END_C

            formatted_record = TermColor.WHITE + str(formatted_record) + TermColor.END_C

            formatted_record = re.sub(
                asc_time + "\t" + level_name + "\t" + name + message,
                colored_asc_time + "\t" + colored_level_name + "\t" + colored_name + colored_message,
                formatted_record
            )

            return formatted_record
        except Exception as e:
            # TODO :fix recursion
            # logger.debug(e)
            return Formatter.format(self, record_cpy)


class MaxLevelFilter(logging.Filter):
    """
    Sets a filter to the logging level
    """
    def __init__(self, level):
        super().__init__(name='')
        self.level = level

    def filter(self, record) -> bool:
        """
        filters a given logging record and
        returns True if logging levels falls
        under the specified max logging level

        Args:
            record: logging record to be inspected

        Returns:
            True if logging level is less than max
                logging level, or else False
        """
        return record.levelno < self.level



import json
import logging
from copy import deepcopy
from typing import Any, Dict, List, Text

import rasa.shared.utils.io
import rasa.utils.io
import regex
from rasa.nlu.tokenizers.tokenizer import Token, Tokenizer
from rasa.shared.nlu.training_data.message import Message

from seetm.core.token_mapper import TokenMapper
from seetm.shared.constants import (
    MappingMethod,
    Config,
    InterfaceType,
    DEFAULT_TOKENIZER_PERSIST_PATH, FilePermission, Encoding,
)
from seetm.utils.config import get_init_configs

logger = logging.getLogger(__name__)


class SEETMTokenizer(Tokenizer):
    defaults = {
        # Case sensitivity
        "case_sensitive": True,
        # Persisting mappings
        "persist": False
    }

    # the following language should not be tokenized using the WhitespaceTokenizer
    not_supported_language_list = ["zh", "ja", "th"]

    def __init__(self, component_config: Dict[Text, Any] = None) -> None:
        """Construct a new tokenizer using the WhitespaceTokenizer framework."""

        super().__init__(component_config)

        self.emoji_pattern = rasa.utils.io.get_emoji_regex()
        self.case_sensitive = component_config["case_sensitive"] if "case_sensitive" in component_config else True
        self.persist_mapped = component_config["persist"] if "persist" in component_config else False
        self.seetm_configs = get_init_configs(
            config_path=Config.DEFAULT_CONFIG_PATH,
            method=MappingMethod.RULE_BASED,
            interface=InterfaceType.MAPPER,
            case=self.case_sensitive,
        )
        self.token_mapper = TokenMapper(
            method=MappingMethod.RULE_BASED,
            case_sensitive=self.case_sensitive,
            quiet_mode=True
        )

        if self.case_sensitive in self.component_config:
            rasa.shared.utils.io.raise_warning("All English tokens will be lowercased "
                                               "prior to the mapping process")

        if self.persist_mapped:
            rasa.shared.utils.io.raise_warning("Persisting mapped instances may affect "
                                               "the training duration of the model")
            try:
                seetm_tokenizer_mappings = {
                    "mapped_instances": []
                }
                with open(
                        DEFAULT_TOKENIZER_PERSIST_PATH,
                        encoding=Encoding.UTF8,
                        mode=FilePermission.WRITE
                ) as seetm_tokenizer_mappings_file:
                    json.dump(
                        seetm_tokenizer_mappings,
                        seetm_tokenizer_mappings_file,
                        ensure_ascii=False,
                        indent=4
                    )
            except Exception as e:
                logger.error("Error occurred while initializing the mapping persisting source. "
                             "Training will be resumed without persisting maps.")
                logger.debug(e)
                self.persist_mapped = False

    def remove_emoji(self, text: Text) -> Text:
        """Remove emoji if the full text, aka token, matches the emoji regex."""
        match = self.emoji_pattern.fullmatch(text)

        if match is not None:
            return ""

        return text

    def tokenize(self, message: Message, attribute: Text) -> List[Token]:
        text = message.get(attribute)
        previous_text = deepcopy(text)

        # SEETM Mapping
        text = self.token_mapper.map(data_instance=text)
        
        # Persisting mapped instances if requested
        try:
            if self.persist_mapped:
                with open(
                        file=DEFAULT_TOKENIZER_PERSIST_PATH,
                        mode=FilePermission.READ,
                        encoding=Encoding.UTF8
                ) as seetm_tokenizer_mappings_file:
                    seetm_tokenizer_mappings = json.load(fp=seetm_tokenizer_mappings_file)

                seetm_tokenizer_mappings["mapped_instances"] = [
                    *seetm_tokenizer_mappings["mapped_instances"],
                    {
                        "instance": previous_text,
                        "mapped_instance": text
                    }
                ]
                with open(
                        DEFAULT_TOKENIZER_PERSIST_PATH,
                        encoding=Encoding.UTF8,
                        mode=FilePermission.WRITE
                ) as seetm_tokenizer_mappings_file:
                    json.dump(
                        seetm_tokenizer_mappings,
                        seetm_tokenizer_mappings_file,
                        ensure_ascii=False,
                        indent=4
                    )
        except Exception as e:
            logger.debug("Error occurred while persisting mapped instances. "
                         f"Training will be resumed without persisting maps. {e}")

        # we need to use regex instead of re, because of
        # https://stackoverflow.com/questions/12746458/python-unicode-regular-expression-matching-failing-with-some-unicode-characters

        # remove 'not a word character' if
        words = regex.sub(
            # there is a space or an end of a string after it
            r"[^\w#@&]+(?=\s|$)|"
            # there is a space or beginning of a string before it
            # not followed by a number
            r"(\s|^)[^\w#@&]+(?=[^0-9\s])|"
            # not in between numbers and not . or @ or & or - or #
            # e.g. 10'000.00 or blabla@gmail.com
            # and not url characters
            r"(?<=[^0-9\s])[^\w._~:/?#\[\]()@!$&*+,;=-]+(?=[^0-9\s])",
            " ",
            text,
        ).split()

        words = [self.remove_emoji(w) for w in words]
        words = [w for w in words if w]

        # if we removed everything like smiles `:)`, use the whole text as 1 token
        if not words:
            words = [text]

        tokens = self._convert_words_to_tokens(words, text)
        return self._apply_token_pattern(tokens)



import logging
from time import time

from seetm.core.token_mapper import TokenMapper
from seetm.shared.constants import MappingMethod

logger = logging.getLogger(__name__)
tm = TokenMapper(method=MappingMethod.RULE_BASED)


if __name__ == "__main__":
    logger.debug("Started testing the mapper core...")

    text = "castle attitude construct component ConStiTute We are a leading non-state degree awarding " \
           "institute approved by the obesity University Grants Commission (UGC) under the Universities " \
           "Act. We are also members of the A an aviation Association of Commonwealth Universities " \
           "(ACU), as well as the International Association of \"Intel internet calm intercom and the " \
           "first Sri Lankan institute to be accredited by the Institution of Engineering & Technology, UK.\"" \
           "cow love hawk rule bull cool bloom page apple ant degree banana cry requirements dean list " \
           "viva easter barrier Career – In your professional or personal life, you will always be " \
           "asked about your qualifications. If you want to get a good job, then you are supposed " \
           "to have great knowledge in that particular domain, and that is only possible if you have " \
           "acquired a vast amount of knowledge"

    text2 = "SLIIT එකේ තියෙන degrees මොනවද?"

    for t in [text, text2]:
        st = time()
        tm.map(
            data_instance=t,
        )
        et = time()
    print(f"Duration: {et - st}")



import json
import logging
import os
import platform
import signal
import subprocess
from datetime import datetime

from flask import (
    send_from_directory,
    request,
    send_file,
)
from flask_cors import cross_origin

from seetm.core.extractor import SEETMExtractor
from seetm.core.token_mapper import TokenMapper
from seetm.server.seetm_api import blueprint
from seetm.server.seetm_api.utils.server_configs import ServerConfigs
from seetm.server.seetm_api.utils.statistics import (
    model_statistics,
    explanation_statistics
)
from seetm.shared.constants import (
    Validity,
    ServerEnv,
    ServerConfigType, Config, DEFAULT_DATA_PATH, InterfaceType, TOKEN_TO_TOKEN_MAP_PATH, FilePermission, Encoding,
)
from seetm.shared.exceptions.server import InvalidServerConfigsException
from seetm.utils.config import get_init_configs
from seetm.utils.io import file_exists

logger = logging.getLogger(__name__)


@blueprint.route("/", strict_slashes=False, methods=['GET'])
@blueprint.route("/status", strict_slashes=False, methods=['GET'])
@cross_origin()
def api_status():
    logger.debug("Status API endpoint was called")
    return {
        "endpoint": "/api/seetm/",
        "status": "ok",
        "environment": "production",
    }


# @blueprint.route("/explain", strict_slashes=False, methods=['POST'])
# @cross_origin()
# def explain():
#     logger.debug("Explain API endpoint was called")
#     request_id_for_exception_handling = None
#     process_q = process_queue.ProcessQueue(data_source_path=PROCESS_QUEUE)
#
#     try:
#         request_data = request.get_json()
#         app_env = request_data['app_env']
#
#         # # handle these from frontend side
#         # handle rest type modes
#         # warn about global outputs
#         model_type = request_data['model_type']
#         data_instance = request_data['data_instance']
#         output_mode = 'dual'
#
#         if app_env == ServerEnv.STRICT_LOCAL:
#             if model_type == MODEL_TYPE_DIET:
#                 rasa_dime_explainer = RasaDIMEExplainer(
#                     models_path=request_data['models_path'],
#                     model_name=request_data['model_name'],
#                     testing_data_path=request_data['data_path'],
#                     model_mode=request_data['model_mode'],
#                     rasa_version=request_data['rasa_version'],
#                     url=request_data['url_endpoint'],
#                     data_instances=[data_instance],
#                     ranking_length=request_data['ranking_length'],
#                     case_sensitive=request_data['case_sensitive'],
#                     metric=request_data['metric'],
#                     output_mode=output_mode,
#                 )
#                 explanation_ = rasa_dime_explainer.explain(inspect=True)
#                 return {"status": "success", "explanation": explanation_.get_explanation(ExplanationType.QUIET)}, 200
#             elif model_type == MODEL_TYPE_OTHER:
#                 return {"status": "error", "cause": "Non-DIET models are not supported yet"}, 200
#         else:
#             request_id = request_data['request_id']
#             request_id_for_exception_handling = request_id
#             data_instance = request_data['data_instance']
#
#             # checking if request id is already
#             # present in the process queue
#             process_exists = process_q.check_existence(request_id=request_id)
#             if process_exists:
#                 logger.error("Request Already Exists in the Process Queue")
#                 raise ProcessAlreadyExistsException()
#
#             process_q.push(
#                 process_id=PROCESS_ID_NONE,
#                 request_id=request_id,
#                 timestamp=datetime.now().timestamp(),
#                 metadata=data_instance,
#             )
#             logger.debug(f"Pushed request {request_id} to process queue")
#
#             if platform.system() == "Windows":
#                 sub_p = subprocess.Popen(
#                     ["dime", "explain", "-r", f"{request_id}", "--quiet"],
#                     shell=True,
#                     stderr=subprocess.PIPE,
#                     stdout=subprocess.PIPE,
#                     creationflags=subprocess.CREATE_NEW_PROCESS_GROUP,
#                 )
#             else:
#                 sub_p = subprocess.Popen(
#                     ["dime", "explain", "-r", f"{request_id}", "--quiet"],
#                     shell=True,
#                     stderr=subprocess.PIPE,
#                     stdout=subprocess.PIPE,
#                     preexec_fn=os.setsid
#                 )
#
#             # saving request and process id to process queue
#             process_id = sub_p.pid
#             process_q.update_pid(
#                 process_id=process_id,
#                 request_id=request_id,
#                 timestamp=datetime.now().timestamp()
#             )
#             logger.debug(f"Updated process queue of {request_id} with process id {process_id}")
#
#             # grabbing stdout and stderr
#             out, err = sub_p.communicate()
#             out, err = out.decode("utf-8"), err.decode("utf-8")
#             return_code = sub_p.returncode
#
#             # setting return_code if errors are present
#             explanation_result = ""
#             if return_code == 1 or "error" in err:
#                 return_code = 1
#             else:
#                 return_code = 0
#
#             # retrieving explanation from process metadata
#             if return_code == 0:
#                 explanation_result = process_q.get_metadata(request_id=request_id)
#
#             # removing from process queue after process is executed
#             process_q.remove(
#                 request_id=request_id
#             )
#             logger.debug(f"Process {process_id} under request id "
#                          f"{request_id} was removed from process queue")
#
#             logger.debug(f"Subprocess output: \n{out}")
#             logger.debug(f"Subprocess errors: \n{err if err else 'None'}")
#             logger.debug(f"Subprocess return code: \n{return_code}")
#
#             response_payload = {
#                 "status": "success" if return_code == 0 else "error",
#                 "process_id": process_id,
#                 "request_id": request_id,
#                 "explanation": explanation_result,
#             }
#             return response_payload, 200
#     except Exception as e:
#         # removing the inserted process from the queue
#         if request_id_for_exception_handling:
#             process_q.remove(
#                 request_id=request_id_for_exception_handling
#             )
#         logger.error(f"Exception occurred while generating an Explanation "
#                      f"for request id {request_id_for_exception_handling}. {e}")
#         return {"status": "error"}, 200
#
#
# @blueprint.route("/abort", strict_slashes=False, methods=['POST'])
# @cross_origin()
# def abort():
#     logger.debug("Process abort API endpoint was called")
#
#     try:
#         request_data = request.get_json()
#         process_q = process_queue.ProcessQueue(data_source_path=PROCESS_QUEUE)
#         request_id = request_data["request_id"]
#
#         if not request_id:
#             raise InvalidRequestIDException()
#
#         # check process existence
#         process_exists = process_q.check_existence(request_id=request_id)
#         if not process_exists:
#             logger.error("Removing non-existing processes is not allowed")
#             raise ProcessNotExistsException()
#
#         # retrieving process_id
#         process_id = process_q.get_pid(request_id=request_id)
#         logger.debug(f"Confirmed the existence of the process {process_id} with the request id: {request_id}")
#
#         if platform.system() == "Windows":
#             kill_process_tree(int(process_id))
#         else:
#             os.killpg(os.getpgid(int(process_id)), signal.SIGTERM)
#         logger.debug(f"Removed the existing process {process_id} with the request id: {request_id}")
#
#         # updating the process queue
#         process_q.remove(request_id=request_id)
#         logger.debug(f"Removed process from queue. [process id: {process_id}, request id: {request_id}]")
#         return {"status": "success", "process_id": process_id}, 200
#     except InvalidProcessIDException as e:
#         logger.error(f"Invalid process ID received. {e}")
#         return {"status": "error"}
#     except ProcessTerminationException as e:
#         logger.error(f"Process termination error. {e}")
#         return {"status": "error"}
#     except Exception as e:
#         logger.error(f"Unknown process termination exception {e}")
#         return {"status": "error"}


@blueprint.route("/maps", strict_slashes=False, methods=['GET', 'POST', 'DELETE', 'PUT'])
@cross_origin()
def maps():
    # logger.debug("Maps API endpoint was called")
    # try:
    # mappable_tokens = list()
    # unmappable_tokens = list()
    #
    # try:
    #     # update the existing maps
    #     current_configs = get_init_configs(
    #         config_path=Config.DEFAULT_CONFIG_PATH,
    #         data_path=DEFAULT_DATA_PATH,
    #         interface=InterfaceType.EXTRACTOR,
    #     )
    #     seetm_extractor = SEETMExtractor(
    #         configs=current_configs,
    #         persist=True,
    #         clean=False,
    #         quiet_mode=False,
    #     )
    #     mappable_tokens, unmappable_tokens = seetm_extractor.extract()
    # except Exception:
    #     raise InvalidServerConfigsException()
    #
    # return {'status': 'success', 'tokens': mappable_tokens + unmappable_tokens}

    logger.debug("Maps API endpoint was called")
    if request.method == "GET":
        try:
            with open(
                    file=TOKEN_TO_TOKEN_MAP_PATH,
                    mode=FilePermission.READ,
                    encoding=Encoding.UTF8
            ) as token_to_token_map_file:
                token_to_token_maps = json.load(fp=token_to_token_map_file)

            return {"status": "success", "maps": token_to_token_maps}, 200
        except Exception as e:
            logger.error(f"Exception occurred while retrieving the maps. {e}")
            return {"status": "error"}, 700

    elif request.method == "POST":
        try:
            request_data = request.get_json()
            map_ = request_data['map']

            with open(
                    file=TOKEN_TO_TOKEN_MAP_PATH,
                    mode=FilePermission.READ,
                    encoding=Encoding.UTF8
            ) as token_to_token_map_file:
                token_to_token_maps = json.load(fp=token_to_token_map_file)

            token_to_token_maps.update(map_)
            with open(TOKEN_TO_TOKEN_MAP_PATH, encoding=Encoding.UTF8, mode=FilePermission.WRITE) as updated_map:
                json.dump(token_to_token_maps, updated_map, ensure_ascii=False, indent=4)

            return {"status": "success", "maps": token_to_token_maps}, 200
        except Exception as e:
            logger.error(f"Exception occurred while persisting the map. {e}")
            return {"status": "error"}, 701

    elif request.method == "DELETE":
        try:
            request_data = request.get_json()
            base_token = request_data["base_token"]

            with open(
                    file=TOKEN_TO_TOKEN_MAP_PATH,
                    mode=FilePermission.READ,
                    encoding=Encoding.UTF8
            ) as token_to_token_map_file:
                token_to_token_maps = json.load(fp=token_to_token_map_file)

            if base_token in token_to_token_maps:
                del token_to_token_maps[base_token]

            with open(TOKEN_TO_TOKEN_MAP_PATH, encoding=Encoding.UTF8, mode=FilePermission.WRITE) as updated_map:
                json.dump(token_to_token_maps, updated_map, ensure_ascii=False, indent=4)

            return {"status": "success", "maps": token_to_token_maps}, 200
        except Exception as e:
            logger.exception(f"Exception occurred while deleting the map. {e}")
            return {"status": "error"}, 702

    elif request.method == "PUT":
        try:
            request_data = request.get_json()
            previous_base_token = request_data["previous_base_token"]
            updated_map = request_data["updated_map"]

            with open(
                    file=TOKEN_TO_TOKEN_MAP_PATH,
                    mode=FilePermission.READ,
                    encoding=Encoding.UTF8
            ) as token_to_token_map_file:
                token_to_token_maps = json.load(fp=token_to_token_map_file)

            print(token_to_token_maps)
            if previous_base_token in token_to_token_maps:
                del token_to_token_maps[previous_base_token]

            if list(updated_map.keys())[0] in token_to_token_maps:
                del token_to_token_maps[list(updated_map.keys())[0]]

            token_to_token_maps.update(updated_map)
            with open(TOKEN_TO_TOKEN_MAP_PATH, encoding=Encoding.UTF8, mode=FilePermission.WRITE) as updated_map:
                json.dump(token_to_token_maps, updated_map, ensure_ascii=False, indent=4)

            return {"status": "success", "maps": token_to_token_maps}, 200
        except Exception as e:
            logger.exception(f"Exception occurred while deleting the map. {e}")
            return {"status": "error"}, 703


# @blueprint.route("/model", strict_slashes=False, methods=['DELETE'])
# @cross_origin()
# def model():
#     logger.debug("Models API endpoint was called")
#     if request.method == "DELETE":
#         try:
#             request_data = request.get_json()
#             models_path = request_data["models_path"]
#             model_name = request_data["model_name"]
#             model_path = os.path.join(models_path, model_name)
#
#             if not file_exists(file_path=model_path):
#                 logger.error("Specified model does not exist")
#                 raise ModelNotFoundException()
#
#             os.remove(path=model_path)
#             logger.debug(f"Deleted the specified model {model_name}")
#
#             return {"status": "success", "model_name": model_name}, 200
#         except Exception as e:
#             logger.error(f"Exception occurred while deleting the model. {e}")
#             return {"status": "error"}, 200
#     else:
#         return {"status": "error"}, 200
#
#
# @blueprint.route("/explanation", strict_slashes=False, methods=['GET', 'POST', 'DELETE'])
# @cross_origin()
# def explanation():
#     logger.debug("Explanations API endpoint was called")
#     explanations_path = DEFAULT_PERSIST_PATH
#     if request.method == "DELETE":
#         logger.debug(f"Deleting an explanation...")
#         try:
#             request_data = request.get_json()
#             explanation_name = request_data["explanation_name"]
#             explanation_path = os.path.join(explanations_path, explanation_name)
#
#             if not file_exists(file_path=explanation_path):
#                 logger.error("Specified explanation does not exist")
#                 raise ExplanationNotFoundException()
#
#             os.remove(path=explanation_path)
#             logger.debug(f"Deleted the specified explanation {explanation_name}")
#
#             return {"status": "success", "explanation_name": explanation_name}, 200
#         except Exception as e:
#             logger.error(f"Exception occurred while deleting the explanation. {e}")
#             return {"status": "error"}, 200
#     elif request.method == "POST":
#         logger.debug(f"Uploading an explanation...")
#         try:
#             request_data = request.get_json()
#             explanation_ = request_data["explanation"]
#             explanation_obj = load_explanation(explanation=explanation_)
#             explanation_obj.persist()
#             logger.debug(f"Explanation uploaded as {explanation_obj.file_name}")
#
#             return {"status": "success", "explanation_name": explanation_obj.file_name}, 200
#         except Exception as e:
#             logger.error(f"Exception occurred while uploading the explanation. {e}")
#             return {"status": "error"}, 200
#     else:
#         logger.debug(f"Downloading an explanation...")
#         try:
#             explanation_name = request.args.get("explanation_name", "")
#             if not explanation_name:
#                 raise InvalidExplanationSpecifiedException()
#
#             logger.debug(f"Sending {explanation_name} file...")
#             return send_file(
#                 path_or_file=os.path.join(os.getcwd(), explanations_path, explanation_name),
#                 as_attachment=True
#             )
#         except Exception as e:
#             logger.error(f"Exception occurred while downloading the explanation. {e}")
#             return {"status": "error"}, 200
#
#
# @blueprint.route("/explanation/visualize", strict_slashes=False, methods=["GET"])
# @cross_origin()
# def visualize_explanation():
#     logger.debug("Explanation visualize API endpoint was called")
#     try:
#         explanation_name = request.args.get("explanation_name", "")
#         if not explanation_name:
#             raise InvalidExplanationSpecifiedException()
#
#         explanation_ = load_explanation(explanation=explanation_name)
#         metadata = explanation_.get_explanation(output_type=ExplanationType.QUIET)
#         return {
#                    "status": "success",
#                    "explanation": metadata
#                }, 200
#     except Exception as e:
#         logger.error(f"Exception occurred while getting stats. {e}")
#         return {"status": "error"}, 200


@blueprint.route("/configs", strict_slashes=False, methods=['GET', 'POST'])
@cross_origin()
def configs():
    logger.debug("Configs API endpoint was called")
    server_configs = ServerConfigs()

    if request.method == 'POST':
        return {}
        # try:
        #     request_data = request.get_json()
        #     updated_configs = request_data['updated_configs']
        #     validate_status, validate_res = server_configs.validate(configs=updated_configs)
        #     if not validate_status:
        #         return {"status": Validity.INVALID, "metadata": validate_res}, 200
        #
        #     server_configs.update_and_persist(
        #         updated_configs=updated_configs,
        #         validate=False
        #     )
        #     return {"status": Validity.VALID}, 200
        #
        # except Exception as e:
        #     logger.error(f"Exception occurred while updating server configs. {e}")
        #     return {"status": "error", "validations": None}, 200
    else:
        try:
            server_configs_json = server_configs.retrieve(
                config_type=ServerConfigType.JSON,
                custom_configs=True
            )
            return {"status": "success", "configs": server_configs_json}
        except Exception as e:
            logger.error(f"Exception occurred while retrieving the server configurations. {e}")
            return {"status": "error"}


@blueprint.route("/<path:path>", strict_slashes=False)
@cross_origin()
def api_static_files(path):
    logger.debug("API static files are served")
    if path != 'seetm.png':
        return {"status": "error", "message": "Not Authorized"}, 403
    return send_from_directory(directory=blueprint.static_folder, path=path), 200



import logging
from difflib import SequenceMatcher
from typing import List, Text

from seetm.shared.exceptions.core import (
    InvalidNumberOfTokensException,
)

logger = logging.getLogger(__name__)


def similarity(tokens: List[Text], inspect: bool = False) -> float:
    if len(tokens) != 2:
        raise InvalidNumberOfTokensException()

    token_1 = tokens[0]
    token_2 = tokens[1]

    sequence_matcher = SequenceMatcher(a=token_1, b=token_2)
    similarity_score = sequence_matcher.ratio()

    # displaying the similarity score if required
    if inspect:
        print(similarity_score)
    return similarity_score


if __name__ == "__main__":
    logger.debug("Testing the Sequence similarity score...")
    score = similarity(["dog", "dog"], inspect=True)



import logging

import regex

from seetm.core.maps.phonetics_to_sinhala_maps import (
    SI_VOWEL_KEYS,
    SI_VOWEL_PURE_MAP,
    SI_VOWEL_SUFFIX_MAP,
    SI_CONSONANT_KEYS,
    SI_CONSONANT_MAP,
    SI_NON_JOINING_KEYS,
    SI_NON_JOINING_MAP,
    SI_SPECIAL_KEYS,
    SI_SPECIAL_MAP,
)

logger = logging.getLogger(__name__)


class PhoneticsToSinhalaMapper:
    @staticmethod
    def convert(text):
        # Replacing non-joining characters
        for nj_char in SI_NON_JOINING_KEYS:
            en_pattern = nj_char
            si_pattern = SI_NON_JOINING_MAP[nj_char]
            text = regex.sub(en_pattern, si_pattern, text)

        # Replacing special consonants
        for sp_char in SI_SPECIAL_KEYS:
            for con_char in SI_CONSONANT_KEYS:
                en_pattern = con_char + sp_char
                si_pattern = SI_CONSONANT_MAP[con_char] + SI_SPECIAL_MAP[sp_char]
                text = regex.sub(en_pattern, si_pattern, text)

        # Replacing consonants + Rakaransha + vowel modifiers
        for con_char in SI_CONSONANT_KEYS:
            for v_char in SI_VOWEL_KEYS:
                en_pattern = con_char + "r" + v_char
                si_pattern = SI_CONSONANT_MAP[con_char] + "්‍ර" + SI_VOWEL_SUFFIX_MAP[v_char]
                text = regex.sub(en_pattern, si_pattern, text)

            en_pattern = con_char + "r"
            si_pattern = SI_CONSONANT_MAP[con_char] + "්‍ර"
            text = regex.sub(en_pattern, si_pattern, text)

        # Replacing consonants + vowel modifiers
        for con_char in SI_CONSONANT_KEYS:
            for v_char in SI_VOWEL_KEYS:
                en_pattern = con_char + v_char
                si_pattern = SI_CONSONANT_MAP[con_char] + SI_VOWEL_SUFFIX_MAP[v_char]
                text = regex.sub(en_pattern, si_pattern, text)

        # Replacing consonants + HAL
        for con_char in SI_CONSONANT_KEYS:
            en_pattern = con_char
            si_pattern = SI_CONSONANT_MAP[con_char] + "්"
            text = regex.sub(en_pattern, si_pattern, text)

        # Replacing vowels
        for v_char in SI_VOWEL_KEYS:
            en_pattern = v_char
            si_pattern = SI_VOWEL_PURE_MAP[v_char]
            text = regex.sub(en_pattern, si_pattern, text)

        return text


if __name__ == "__main__":
    logger.debug("Started testing the phonetic mapper...")
    print(PhoneticsToSinhalaMapper.convert(
        """
        dhis iis a test
        """
    ))



import logging
import os.path
import time
from typing import List, Text, Optional

import gensim.models
from gensim.models import Word2Vec
from gensim.models.callbacks import CallbackAny2Vec
from rich.progress import Progress

from seetm.shared.constants import DEFAULT_EMBEDDING_MODELS_PATH
from seetm.shared.exceptions.core import (
    Word2VecModelTrainingException,
    InvalidNumberOfTokensException, Word2VecModelNotFoundException,
)
from seetm.utils.io import get_timestamp_str
from seetm.utils.text_preprocessing import tokenize


logger = logging.getLogger(__name__)
gensim.models.word2vec.logger.level = 60
gensim.utils.logger.level = 60


class Word2VecModel:
    def __init__(self, dataset: List[Text], model_name: Text, epochs: int = 10):
        self.dataset = self._preprocess(dataset)
        self.name = model_name if model_name else get_timestamp_str()
        self.model = None
        self.vocabulary = list()
        self.epochs = epochs
        self.epoch_progress = EpochProgress(total_epochs=epochs, model_type=model_name)

    @staticmethod
    def _preprocess(dataset: List) -> List:
        preprocessed_dataset = [tokenize(str.lower(instance)) for instance in dataset]
        return preprocessed_dataset

    def train(self):
        try:
            self.model = Word2Vec(
                sentences=self.dataset,
                min_count=1,
                callbacks=[self.epoch_progress],
                epochs=self.epochs,
            )
            self.vocabulary = list(self.model.wv.key_to_index.keys())
        except Exception as e:
            raise Word2VecModelTrainingException(e)

    def persist(self):
        self.model.save(os.path.join(DEFAULT_EMBEDDING_MODELS_PATH, self.name + ".w2v"))

    def cosine_similarity(self, instances: List) -> float:
        if len(instances) != 2:
            raise InvalidNumberOfTokensException()

        if not self.model:
            raise Word2VecModelNotFoundException()

        if instances[0] not in self.vocabulary or instances[1] not in self.vocabulary:
            return 0.0

        model: Word2Vec = self.model
        return model.wv.n_similarity(ws1=instances[0].split(), ws2=instances[1].split())

    def similar_tokens(self, token: Text) -> List:
        if not self.model:
            raise Word2VecModelNotFoundException()

        model: Word2Vec = self.model
        return model.wv.most_similar(token)


class EpochProgress(CallbackAny2Vec):
    def __init__(self, total_epochs: int, model_type: Text):
        self.epoch = 0
        self.training_task = None
        self.total_epochs = total_epochs
        self.model_type = model_type
        self.progress = Progress(transient=True)
        self._initialize_progress()

    def _initialize_progress(self):
        # initializing model type
        if self.model_type == "nlu_model":
            self.model_type = "NLU"
        elif self.model_type == "ipa_model":
            self.model_type = "IPA"
        else:
            self.model_type = "Rule-based"

        # initializing progress
        self.training_task = self.progress.add_task(f"Training Word2Vec [{self.model_type}]", total=self.total_epochs)

    def on_epoch_end(self, model):
        self.epoch += 1
        with self.progress as progress:
            progress.update(self.training_task, advance=1)
            time.sleep(0.1)

        if self.epoch == self.total_epochs:
            logger.info(f"{self.model_type} Word2Vec model was successfully "
                        f"trained for a total of {self.total_epochs} epochs")
